import base64
import codecs
import hashlib
import os
import warnings

import numpy as np

import westpa
from westpa.core import h5io
from westpa.cli.tools import w_assign, w_direct, w_reweight

from westpa.tools import WESTParallelTool, WESTDataReader, ProgressIndicatorComponent, Plotter

from westpa.tools import WIPIDataset, __get_data_for_iteration__, WIPIScheme


warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=RuntimeWarning)
warnings.filterwarnings('ignore', category=FutureWarning)
warnings.filterwarnings('ignore', category=ImportWarning)
warnings.filterwarnings('ignore')


class WIPI(WESTParallelTool):
    '''
    Welcome to w_ipa (WESTPA Interactive Python Analysis)!
    From here, you can run traces, look at weights, progress coordinates, etc.
    This is considered a 'stateful' tool; that is, the data you are pulling is always pulled
    from the current analysis scheme and iteration.
    By default, the first analysis scheme in west.cfg is used, and you are set at iteration 1.

    ALL PROPERTIES ARE ACCESSED VIA w or west
    To see the current iteration, try:

        w.iteration
        OR
        west.iteration

    to set it, simply plug in a new value.

        w.iteration = 100

    To change/list the current analysis schemes:

        w.list_schemes
        w.scheme = OUTPUT FROM w.list_schemes

    To see the states and bins defined in the current analysis scheme:

        w.states
        w.bin_labels

    All information about the current iteration is available in an object called 'current':

        w.current
        walkers, summary, states, seg_id, weights, parents, kinavg, pcoord, bins, populations, and auxdata, if it exists.

    In addition, the function w.trace(seg_id) will run a trace over a seg_id in the current iteration and return a dictionary
    containing all pertinent information about that seg_id's history.  It's best to store this, as the trace can be expensive.

    Run help on any function or property for more information!

    Happy analyzing!

    '''

    def __init__(self):
        super().__init__()
        self.data_reader = WESTDataReader()
        self.wm_env.default_work_manager = self.wm_env.default_parallel_work_manager
        self.progress = ProgressIndicatorComponent()

        self._iter = 1
        self.config_required = True
        self.version = "1.0B"
        # Set to matplotlib if you want that.  But why would you?
        # Well, whatever, we'll just set it to that for now.
        self.interface = 'matplotlib'
        self._scheme = None
        global iteration

    def add_args(self, parser):
        self.progress.add_args(parser)
        self.data_reader.add_args(parser)
        rgroup = parser.add_argument_group('runtime options')
        rgroup.add_argument(
            '--analysis-only',
            '-ao',
            dest='analysis_mode',
            action='store_true',
            help='''Use this flag to run the analysis and return to the terminal.''',
        )
        rgroup.add_argument(
            '--reanalyze',
            '-ra',
            dest='reanalyze',
            action='store_true',
            help='''Use this flag to delete the existing files and reanalyze.''',
        )
        rgroup.add_argument(
            '--ignore-hash', '-ih', dest='ignore_hash', action='store_true', help='''Ignore hash and don't regenerate files.'''
        )
        rgroup.add_argument(
            '--debug', '-d', dest='debug_mode', action='store_true', help='''Debug output largely intended for development.'''
        )
        rgroup.add_argument('--terminal', '-t', dest='plotting', action='store_true', help='''Plot output in terminal.''')
        # There is almost certainly a better way to handle this, but we'll sort that later.
        import argparse

        rgroup.add_argument('--f', '-f', dest='extra', default='blah', help=argparse.SUPPRESS)

        parser.set_defaults(compression=True)

    def process_args(self, args):
        self.progress.process_args(args)
        self.data_reader.process_args(args)
        with self.data_reader:
            self.niters = self.data_reader.current_iteration - 1
        self.__config = westpa.rc.config
        self.__settings = self.__config['west']['analysis']
        for ischeme, scheme in enumerate(self.__settings['analysis_schemes']):
            if (
                self.__settings['analysis_schemes'][scheme]['enabled'] is True
                or self.__settings['analysis_schemes'][scheme]['enabled'] is None
            ):
                self.scheme = scheme
        self.data_args = args
        self.analysis_mode = args.analysis_mode
        self.reanalyze = args.reanalyze
        self.ignore_hash = args.ignore_hash
        self.debug_mode = args.debug_mode
        if args.plotting:
            self.interface = 'text'

    def hash_args(self, args, extra=None, path=None):
        '''Create unique hash stamp to determine if arguments/file is different from before.'''
        '''Combine with iteration to know whether or not file needs updating.'''
        # Why are we not loading this functionality into the individual tools?
        # While it may certainly be useful to store arguments (and we may well do that),
        # it's rather complex and nasty to deal with pickling and hashing arguments through
        # the various namespaces.
        # In addition, it's unlikely that the functionality is desired at the individual tool level,
        # since we'll always just rewrite a file when we call the function.
        # return hashlib.md5(pickle.dumps([args, extra])).hexdigest()
        # We don't care about the path, so we'll remove it.
        # Probably a better way to do this, but who cares.
        cargs = list(args)
        for iarg, arg in enumerate(cargs):
            if path in arg:
                cargs[iarg] = arg.replace(path, '').replace('/', '')
            if arg == '--disable-averages':
                cargs.remove('--disable-averages')
        to_hash = cargs + [extra]
        # print(args)
        # print(to_hash)
        # print(str(to_hash).encode('base64'))
        if self.debug_mode:
            for iarg, arg in enumerate(to_hash):
                if not isinstance(arg, list):
                    print('arg {num:02d} -- {arg:<20}'.format(num=iarg, arg=arg))
                else:
                    for il, l in enumerate(arg):
                        print('arg {num:02d} -- {arg:<20}'.format(num=il + iarg, arg=l))
            # print('args: {}'.format(to_hash))
        # This SHOULD produce the same output, maybe?  That would be nice, anyway.
        # But we'll need to test it more.
        return hashlib.md5(base64.b64encode(str(to_hash).encode())).hexdigest()

    def stamp_hash(self, h5file_name, new_hash):
        '''Loads a file, stamps it, and returns the opened file in read only'''
        h5file = h5io.WESTPAH5File(h5file_name, 'r+')
        h5file.attrs['arg_hash'] = new_hash
        h5file.close()
        h5file = h5io.WESTPAH5File(h5file_name, 'r')
        return h5file

    def analysis_structure(self):
        '''
        Run automatically on startup.  Parses through the configuration file, and loads up all the data files from the different
        analysis schematics.  If they don't exist, it creates them automatically by hooking in to existing analysis routines
        and going from there.

        It does this by calling in the make_parser_and_process function for w_{assign,reweight,direct} using a custom built list
        of args.  The user can specify everything in the configuration file that would have been specified on the command line.

        For instance, were one to call w_direct as follows:

            w_direct --evolution cumulative --step-iter 1 --disable-correl

        the west.cfg would look as follows:

        west:
          analysis:
            w_direct:
              evolution: cumulative
              step_iter: 1
              extra: ['disable-correl']

        Alternatively, if one wishes to use the same options for both w_direct and w_reweight, the key 'w_direct' can be replaced
        with 'kinetics'.
        '''
        # Make sure everything exists.
        try:
            os.mkdir(self.__settings['directory'])
        except Exception:
            pass
        # Now, check to see whether they exist, and then load them.
        self.__analysis_schemes__ = {}
        # We really need to implement some sort of default behavior if an analysis scheme isn't set.
        # Right now, we just crash.  That isn't really graceful.
        for scheme in self.__settings['analysis_schemes']:
            if self.__settings['analysis_schemes'][scheme]['enabled']:
                if self.work_manager.running is False:
                    self.work_manager.startup()
                path = os.path.join(os.getcwd(), self.__settings['directory'], scheme)
                # if 'postanalysis' in self.__settings['analysis_schemes'][scheme] and 'postanalysis' in self.__settings['postanalysis']:
                # Should clean this up.  But it uses the default global setting if a by-scheme one isn't set.
                if 'postanalysis' in self.__settings:
                    if 'postanalysis' in self.__settings['analysis_schemes'][scheme]:
                        pass
                    else:
                        self.__settings['analysis_schemes'][scheme]['postanalysis'] = self.__settings['postanalysis']
                try:
                    os.mkdir(path)
                except Exception:
                    pass
                self.__analysis_schemes__[scheme] = {}
                try:
                    if (
                        self.__settings['analysis_schemes'][scheme]['postanalysis'] is True
                        or self.__settings['postanalysis'] is True
                    ):
                        analysis_files = ['assign', 'direct', 'reweight']
                    else:
                        analysis_files = ['assign', 'direct']
                except Exception:
                    analysis_files = ['assign', 'direct']
                    self.__settings['analysis_schemes'][scheme]['postanalysis'] = False
                reanalyze_kinetics = False
                assign_hash = None
                for name in analysis_files:
                    arg_hash = None
                    if self.reanalyze is True:
                        reanalyze_kinetics = True
                        try:
                            os.remove(os.path.join(path, '{}.h5'.format(name)))
                        except Exception:
                            pass
                    else:
                        try:
                            # Try to load the hash.  If we fail to load the hash or the file, we need to reload.
                            # if self.reanalyze == True:
                            #    raise ValueError('Reanalyze set to true.')
                            self.__analysis_schemes__[scheme][name] = h5io.WESTPAH5File(
                                os.path.join(path, '{}.h5'.format(name)), 'r'
                            )
                            arg_hash = self.__analysis_schemes__[scheme][name].attrs['arg_hash']
                            if name == 'assign':
                                assign_hash = arg_hash
                        except Exception:
                            pass
                            # We shouldn't rely on this.
                            # self.reanalyze = True
                    if True:
                        if name == 'assign':
                            assign = w_assign.WAssign()

                            w_assign_config = {'output': os.path.join(path, '{}.h5'.format(name))}
                            try:
                                w_assign_config.update(self.__settings['w_assign'])
                            except Exception:
                                pass
                            try:
                                w_assign_config.update(self.__settings['analysis_schemes'][scheme]['w_assign'])
                            except Exception:
                                pass
                            args = []
                            for key, value in w_assign_config.items():
                                if key != 'extra':
                                    args.append(str('--') + str(key).replace('_', '-'))
                                    args.append(str(value))
                            # This is for stuff like disabling correlation analysis, etc.
                            if 'extra' in list(w_assign_config.keys()):
                                # We're sorting to ensure that the order doesn't matter.
                                for value in sorted(w_assign_config['extra']):
                                    args.append(str('--') + str(value).replace('_', '-'))
                            # We're just calling the built in function.
                            # This is a lot cleaner than what we had in before, and far more workable.
                            args.append('--config-from-file')
                            args.append('--scheme-name')
                            args.append('{}'.format(scheme))
                            # Why are we calling this if we're not sure we're remaking the file?
                            # We need to load up the bin mapper and states and see if they're the same.
                            assign.make_parser_and_process(args=args)
                            import pickle

                            # new_hash = self.hash_args(args=args, path=path, extra=[self.niters, pickle.dumps(assign.binning.mapper), assign.states])
                            # We need to encode it properly to ensure that some OS specific thing doesn't kill us.  Same goes for the args, ultimately.
                            # Mostly, we just need to ensure that we're consistent.
                            new_hash = self.hash_args(
                                args=args,
                                path=path,
                                extra=[
                                    int(self.niters),
                                    codecs.encode(pickle.dumps(assign.binning.mapper), "base64"),
                                    base64.b64encode(str(assign.states).encode()),
                                ],
                            )
                            # Let's check the hash.  If the hash is the same, we don't need to reload.
                            if self.debug_mode is True:
                                print('{:<10}: old hash, new hash -- {}, {}'.format(name, arg_hash, new_hash))
                            if self.ignore_hash is False and (arg_hash != new_hash or self.reanalyze is True):
                                # If the hashes are different, or we need to reanalyze, delete the file.
                                try:
                                    os.remove(os.path.join(path, '{}.h5'.format(name)))
                                except Exception:
                                    pass
                                print('Reanalyzing file {}.h5 for scheme {}.'.format(name, scheme))
                                # reanalyze_kinetics = True
                                # We want to use the work manager we have here.  Otherwise, just let the tool sort out what it needs, honestly.
                                assign.work_manager = self.work_manager

                                assign.go()
                                assign.data_reader.close()

                                # Stamp w/ hash, then reload as read only.
                                self.__analysis_schemes__[scheme][name] = self.stamp_hash(
                                    os.path.join(path, '{}.h5'.format(name)), new_hash
                                )
                            del assign
                            # Update the assignment hash.
                            assign_hash = new_hash

                        # Since these are all contained within one tool, now, we want it to just... load everything.
                        if name == 'direct' or name == 'reweight':
                            if name == 'direct':
                                analysis = w_direct.WDirect()
                            if name == 'reweight':
                                analysis = w_reweight.WReweight()

                            analysis_config = {
                                'assignments': os.path.join(path, '{}.h5'.format('assign')),
                                'output': os.path.join(path, '{}.h5'.format(name)),
                                'kinetics': os.path.join(path, '{}.h5'.format(name)),
                            }

                            # Pull from general analysis options, then general SPECIFIC options for each analysis,
                            # then general options for that analysis scheme, then specific options for the analysis type in the scheme.

                            try:
                                analysis_config.update(self.__settings['kinetics'])
                            except Exception:
                                pass
                            try:
                                analysis_config.update(self.__settings['w_{}'.format(name)])
                            except Exception:
                                pass
                            try:
                                analysis_config.update(self.__settings['analysis_schemes'][scheme]['kinetics'])
                            except Exception:
                                pass
                            try:
                                analysis_config.update(self.__settings['analysis_schemes'][scheme]['w_{}'.format(name)])
                            except Exception:
                                pass

                            # We're pulling in a default set of arguments, then updating them with arguments from the west.cfg file, if appropriate, after setting the appropriate command
                            # Then, we call the magic function 'make_parser_and_process' with the arguments we've pulled in.
                            # The tool has no real idea it's being called outside of its actual function, and we're good to go.
                            args = ['all']
                            for key, value in analysis_config.items():
                                if key != 'extra':
                                    args.append(str('--') + str(key).replace('_', '-'))
                                    args.append(str(value))
                            # This is for stuff like disabling correlation analysis, etc.
                            if 'extra' in list(analysis_config.keys()):
                                for value in sorted(analysis_config['extra']):
                                    args.append(str('--') + str(value).replace('_', '-'))
                            # We want to not display the averages, so...
                            args.append('--disable-averages')
                            new_hash = self.hash_args(args=args, path=path, extra=[int(self.niters), assign_hash])
                            # if arg_hash != new_hash or self.reanalyze == True or reanalyze_kinetics == True:
                            if self.debug_mode is True:
                                print('{:<10}: old hash, new hash -- {}, {}'.format(name, arg_hash, new_hash))
                            if self.ignore_hash is False and (arg_hash != new_hash or reanalyze_kinetics is True):
                                try:
                                    os.remove(os.path.join(path, '{}.h5'.format(name)))
                                except Exception:
                                    pass
                                print('Reanalyzing file {}.h5 for scheme {}.'.format(name, scheme))
                                analysis.make_parser_and_process(args=args)
                                # We want to hook into the existing work manager.
                                analysis.work_manager = self.work_manager

                                analysis.go()

                                # Open!
                                self.__analysis_schemes__[scheme][name] = self.stamp_hash(
                                    os.path.join(path, '{}.h5'.format(name)), new_hash
                                )
                            del analysis

        # Make sure this doesn't get too far out, here.  We need to keep it alive as long as we're actually analyzing things.
        # self.work_manager.shutdown()
        print("")
        print("Complete!")

    @property
    def assign(self):
        return self.__analysis_schemes__[str(self.scheme)]['assign']

    @property
    def direct(self):
        """
        The output from w_kinavg.py from the current scheme.
        """
        return self.__analysis_schemes__[str(self.scheme)]['direct']

    @property
    def state_labels(self):
        print("State labels and definitions!")
        for istate, state in enumerate(self.assign['state_labels']):
            print('{}: {}'.format(istate, state))
        print('{}: {}'.format(istate + 1, 'Unknown'))

    @property
    def bin_labels(self):
        print("Bin definitions! ")
        for istate, state in enumerate(self.assign['bin_labels']):
            print('{}: {}'.format(istate, state))

    @property
    def west(self):
        return self.data_reader.data_manager.we_h5file

    @property
    def reweight(self):
        if self.__settings['analysis_schemes'][str(self.scheme)]['postanalysis'] is True:
            return self.__analysis_schemes__[str(self.scheme)]['reweight']
        else:
            value = "This sort of analysis has not been enabled."
            current = {
                'bin_prob_evolution': value,
                'color_prob_evolution': value,
                'conditional_flux_evolution': value,
                'rate_evolution': value,
                'state_labels': value,
                'state_prob_evolution': value,
            }
            current.update({'bin_populations': value, 'iterations': value})
            return current

    @property
    def scheme(self):
        '''
        Returns and sets what scheme is currently in use.
        To see what schemes are available, run:

            w.list_schemes

        '''
        # Let's do this a few different ways.
        # We want to return things about the DIFFERENT schemes, if possible.
        if self._scheme is None:
            self._scheme = WIPIScheme(
                scheme=self.__analysis_schemes__, name=self._schemename, parent=self, settings=self.__settings
            )

        # This just ensures that when we call it, it's clean.
        self._scheme.name = None
        return self._scheme

    @scheme.setter
    def scheme(self, scheme):
        self._future = None
        self._current = None
        self._past = None
        if scheme in self.__settings['analysis_schemes']:
            pass
        else:
            for ischeme, schemename in enumerate(self.__settings['analysis_schemes']):
                if ischeme == scheme:
                    scheme = schemename
        if (
            self.__settings['analysis_schemes'][scheme]['enabled'] is True
            or self.__settings['analysis_schemes'][scheme]['enabled'] is None
        ):
            self._schemename = scheme
        else:
            print("Scheme cannot be changed to scheme: {}; it is not enabled!".format(scheme))

    @property
    def list_schemes(self):
        '''
        Lists what schemes are configured in west.cfg file.
        Schemes should be structured as follows, in west.cfg:

        west:
          system:
            analysis:
              directory: analysis
              analysis_schemes:
                scheme.1:
                  enabled: True
                  states:
                    - label: unbound
                      coords: [[7.0]]
                    - label: bound
                      coords: [[2.7]]
                  bins:
                    - type: RectilinearBinMapper
                      boundaries: [[0.0, 2.80, 7, 10000]]
        '''
        # print("The following schemes are available:")
        # print("")
        # for ischeme, scheme in enumerate(self.__settings['analysis_schemes']):
        #    print('{}. Scheme: {}'.format(ischeme, scheme))
        # print("")
        # print("Set via name, or via the index listed.")
        # print("")
        # print("Current scheme: {}".format(self.scheme))
        self._scheme.list_schemes

    @property
    def iteration(self):
        '''
        Returns/sets the current iteration.
        '''
        # print("The current iteration is {}".format(self._iter))
        return self._iter

    @iteration.setter
    def iteration(self, value):
        print("Setting iteration to iter {}.".format(value))
        if value <= 0:
            print("Iteration must begin at 1.")
            value = 1
        if value > self.niters:
            print("Cannot go beyond {} iterations!".format(self.niters))
            print("Setting to {}".format(self.niters))
            value = self.niters
        # We want to trigger a rebuild on our current/past/future bits.
        # The scheme should automatically reset to the proper iteration, but
        # future needs to be manually triggered.
        self._iter = value
        self._future = None
        return self._iter

    @property
    def current(self):
        '''
        The current iteration.  See help for __get_data_for_iteration__
        '''
        return self.scheme[self.scheme.scheme].current

    @property
    def past(self):
        '''
        The previous iteration.  See help for __get_data_for_iteration__
        '''
        return self.scheme[self.scheme.scheme].past

    def trace(self, seg_id):
        '''
        Runs a trace on a seg_id within the current iteration, all the way back to the beginning,
        returning a dictionary containing all interesting information:

            seg_id, pcoord, states, bins, weights, iteration, auxdata (optional)

        sorted in chronological order.


        Call with a seg_id.
        '''
        if seg_id >= self.current.walkers:
            print("Walker seg_id # {} is beyond the max count of {} walkers.".format(seg_id, self.current.walkers))
            return 1
        pi = self.progress.indicator
        with pi:
            pi.new_operation('Tracing scheme:iter:seg_id {}:{}:{}'.format(self.scheme, self.iteration, seg_id), self.iteration)
            current = {'seg_id': [], 'pcoord': [], 'states': [], 'weights': [], 'iteration': [], 'bins': []}
            keys = []
            try:
                current['auxdata'] = {}
                for key in list(self.current['auxdata'].keys()):
                    current['auxdata'][key] = []
                    key = []
            except Exception:
                pass
            for iter in reversed(list(range(1, self.iteration + 1))):
                iter_group = self.data_reader.get_iter_group(iter)
                current['pcoord'].append(iter_group['pcoord'][seg_id, :, :])
                current['states'].append(self.assign['trajlabels'][iter - 1, seg_id, :])
                current['bins'].append(self.assign['assignments'][iter - 1, seg_id, :])
                current['seg_id'].append(seg_id)
                current['weights'].append(iter_group['seg_index']['weight'][seg_id])
                current['iteration'].append(iter)
                try:
                    for key in keys:
                        current['auxdata'][key].append(iter_group['auxdata'][key][seg_id])
                except Exception:
                    pass
                seg_id = iter_group['seg_index']['parent_id'][seg_id]
                if seg_id < 0:
                    # Necessary for steady state simulations.  This means they started in that iteration.
                    break
                pi.progress += 1
        current['seg_id'] = list(reversed(current['seg_id']))
        current['iteration'] = list(reversed(current['iteration']))
        current['states'] = np.concatenate(np.array(list(reversed(current['states']))))
        current['bins'] = np.concatenate(np.array(list(reversed(current['bins']))))
        current['weights'] = np.array(list(reversed(current['weights'])))
        current['pcoord'] = np.concatenate(np.array(list(reversed(current['pcoord']))))
        try:
            for key in keys():
                current['auxdata'][key] = np.concatenate(np.array(list(reversed(current['auxdata'][key]))))
        except Exception:
            pass
        current['state_labels'] = self.assign['state_labels']
        for i in ['pcoord', 'states', 'bins', 'weights']:
            current[i] = WIPIDataset(raw=current[i], key=i)
            if i == 'weights':
                current[i].plotter = Plotter(
                    np.log10(current[i].raw), str('log10 of ' + str(i)), iteration=current[i].raw.shape[0], interface=self.interface
                )
            else:
                current[i].plotter = Plotter(current[i].raw, i, iteration=current[i].raw.shape[0], interface=self.interface)
            current[i].plot = current[i].plotter.plot
        return WIPIDataset(raw=current, key=seg_id)

    @property
    def future(self, value=None):
        '''
        Similar to current/past, but keyed differently and returns different datasets.
        See help for Future.
        '''
        if self._future is None:
            self._future = self.Future(raw=self.__get_children__(), key=None)
            self._future.iteration = self.iteration + 1
        return self._future

    class Future(WIPIDataset):

        # This isn't a real fancy one.
        def __getitem__(self, value):
            if isinstance(value, str):
                print(list(self.__dict__.keys()))
                try:
                    return self.__dict__['raw'][value]
                except Exception:
                    print('{} is not a valid data structure.'.format(value))
            elif isinstance(value, int) or isinstance(value, np.int64):
                # Otherwise, we assume they're trying to index for a seg_id.
                # if value < self.parent.walkers:
                current = {}
                current['pcoord'] = self.__dict__['raw']['pcoord'][value]
                current['states'] = self.__dict__['raw']['states'][value]
                current['bins'] = self.__dict__['raw']['bins'][value]
                current['parents'] = self.__dict__['raw']['parents'][value]
                current['seg_id'] = self.__dict__['raw']['seg_id'][value]
                current['weights'] = self.__dict__['raw']['weights'][value]
                try:
                    current['auxdata'] = {}
                    for key in list(self.__dict__['raw']['auxdata'].keys()):
                        current['auxdata'][key] = self.__dict__['raw']['auxdata'][key][value]
                except Exception:
                    pass
                current = WIPIDataset(current, 'Segment {} in Iter {}'.format(value, self.iteration))
                return current

    def __get_children__(self):
        '''
        Returns all information about the children of a given walker in the current iteration.
        Used to generate and create the future object, if necessary.
        '''

        if self.iteration == self.niters:
            print("Currently at iteration {}, which is the max.  There are no children!".format(self.iteration))
            return 0
        iter_data = __get_data_for_iteration__(value=self.iteration + 1, parent=self)
        future = {
            'weights': [],
            'pcoord': [],
            'parents': [],
            'summary': iter_data['summary'],
            'seg_id': [],
            'walkers': iter_data['walkers'],
            'states': [],
            'bins': [],
        }
        for seg_id in range(0, self.current.walkers):
            children = np.where(iter_data['parents'] == seg_id)[0]
            if len(children) == 0:
                error = "No children for seg_id {}.".format(seg_id)
                future['weights'].append(error)
                future['pcoord'].append(error)
                future['parents'].append(error)
                future['seg_id'].append(error)
                future['states'].append(error)
                future['bins'].append(error)
            else:
                # Now, we're gonna put them in the thing.
                value = self.iteration + 1
                future['weights'].append(iter_data['weights'][children])
                future['pcoord'].append(iter_data['pcoord'][...][children, :, :])
                try:
                    aux_data = iter_data['auxdata'][...][children, :, :]
                    try:
                        future['aux_data'].append(aux_data)
                    except Exception:
                        future['aux_data'] = aux_data
                except Exception:
                    pass
                future['parents'].append(iter_data['parents'][children])
                future['seg_id'].append(iter_data['seg_id'][children])
                future['states'].append(self.assign['trajlabels'][value - 1, children, :])
                future['bins'].append(self.assign['assignments'][value - 1, children, :])
        return future

    def go(self):
        '''
        Function automatically called by main() when launched via the command line interface.
        Generally, call main, not this function.
        '''
        w = self

        print("")
        print("Welcome to w_ipa (WESTPA Interactive Python Analysis) v. {}!".format(w.version))
        print("Run w.introduction for a more thorough introduction, or w.help to see a list of options.")
        print("Running analysis & loading files.")
        self.data_reader.open()
        self.analysis_structure()
        # Seems to be consistent with other tools, such as w_assign.  For setting the iterations.
        self.data_reader.open()
        self.niters = self.data_reader.current_iteration - 1
        self.iteration = self.niters
        try:
            print('Your current scheme, system and iteration are : {}, {}, {}'.format(w.scheme, os.getcwd(), w.iteration))
        except Exception:
            pass

    @property
    def introduction(self):
        '''
        Just spits out an introduction, in case someone doesn't call help.
        '''
        help_string = '''
        Call as a dictionary item or a .attribute:

        w.past, w.current, w.future:

            {current}

        Raw schemes can be accessed as follows:

            w.scheme.{scheme_keys}

            and contain mostly the same datasets associated with w.

        The following give raw access to the h5 files associated with the current scheme

        w.west
        w.assign
        w.direct
        w.reweight

        OTHER:

        {w}

        '''.format(
            current=self.__format_keys__(self.current.__dir__(), split=' ', offset=12),
            scheme_keys=self.__format_keys__(list(self._scheme.raw.keys())),
            w=self.__format_keys__(self.__dir__(), offset=8, max_length=0, split='', prepend='w.'),
        )
        print(help_string)

    # Just a little function to be used with the introduction.
    def __format_keys__(self, keys, split='/', offset=0, max_length=80, prepend=''):
        rtn = ''
        run_length = 0
        for key in keys:
            rtn += prepend + str(key) + split
            run_length += len(str(key))
            if run_length >= max_length:
                run_length = offset
                rtn += '\n' + ' ' * offset
        if rtn[-1] == split:
            return rtn[:-1]
        else:
            return rtn

    @property
    def help(self):
        '''Just a minor function to call help on itself.  Only in here to really help someone get help.'''
        help(self)

    def _repr_pretty_(self, p, cycle):
        self.introduction
        return " "

    def __dir__(self):
        return_list = ['past', 'current', 'future']
        # For the moment, don't expose direct, reweight, or assign, as these are scheme dependent files.
        # They do exist, and always link to the current scheme, however.
        return_list += ['iteration', 'niters', 'scheme', 'list_schemes', 'bin_labels', 'state_labels', 'west', 'trace']
        return sorted(set(return_list))


def entry_point():
    west = WIPI()
    w = west
    # We're gonna print some defaults.
    w.main()
    if w.analysis_mode is False:
        from IPython import embed
        import IPython

        # We're using this to set magic commands.
        # Mostly, we're using it to allow tab completion of objects stored in dictionaries.
        try:
            # Worked on MacOS.  Probably just an older version.
            c = IPython.Config()
        except Exception:
            # Seems to be necessary on Linux, and likely on newer installs.
            c = IPython.terminal.ipapp.load_default_config()
        c.IPCompleter.greedy = True
        embed(banner1='', exit_msg='Leaving w_ipa... goodbye.', config=c)
    print("")


if __name__ == '__main__':
    entry_point()



import warnings

import numpy as np

from westpa.core.data_manager import weight_dtype
from westpa.core import h5io

# From w_kinetics.
from westpa.tools.dtypes import ed_list_dtype
from westpa.core.binning import index_dtype
from westpa.core.kinetics._kinetics import _fast_transition_state_copy  # @UnresolvedImport
from westpa.core.kinetics import find_macrostate_transitions


warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=RuntimeWarning)
warnings.filterwarnings('ignore', category=FutureWarning)


# The old w_kinetics
class WKinetics:
    def w_kinetics(self):
        pi = self.progress.indicator
        pi.new_operation('Initializing')

        self.data_reader.open('r')
        self.open_files()
        nstates = self.assignments_file.attrs['nstates']
        start_iter, stop_iter = self.iter_range.iter_start, self.iter_range.iter_stop  # h5io.get_iter_range(self.assignments_file)
        iter_count = stop_iter - start_iter
        durations_ds = self.output_file.replace_dataset(
            'durations',
            shape=(iter_count, 0),
            maxshape=(iter_count, None),
            dtype=ed_list_dtype,
            chunks=(1, 15360) if self.do_compression else None,
            shuffle=self.do_compression,
            compression=9 if self.do_compression else None,
        )
        durations_count_ds = self.output_file.replace_dataset(
            'duration_count', shape=(iter_count,), dtype=np.int_, shuffle=True, compression=9
        )
        cond_fluxes_ds = self.output_file.replace_dataset(
            'conditional_fluxes',
            shape=(iter_count, nstates, nstates),
            dtype=weight_dtype,
            chunks=(h5io.calc_chunksize((iter_count, nstates, nstates), weight_dtype) if self.do_compression else None),
            shuffle=self.do_compression,
            compression=9 if self.do_compression else None,
        )
        total_fluxes_ds = self.output_file.replace_dataset(
            'total_fluxes',
            shape=(iter_count, nstates),
            dtype=weight_dtype,
            chunks=(h5io.calc_chunksize((iter_count, nstates), weight_dtype) if self.do_compression else None),
            shuffle=self.do_compression,
            compression=9 if self.do_compression else None,
        )

        cond_arrival_counts_ds = self.output_file.replace_dataset(
            'conditional_arrivals',
            shape=(iter_count, nstates, nstates),
            dtype=np.uint,
            chunks=(h5io.calc_chunksize((iter_count, nstates, nstates), np.uint) if self.do_compression else None),
            shuffle=self.do_compression,
            compression=9 if self.do_compression else None,
        )
        arrival_counts_ds = self.output_file.replace_dataset(
            'arrivals',
            shape=(iter_count, nstates),
            dtype=np.uint,
            chunks=(h5io.calc_chunksize((iter_count, nstates), np.uint) if self.do_compression else None),
            shuffle=self.do_compression,
            compression=9 if self.do_compression else None,
        )

        # copy state labels for convenience
        self.output_file.replace_dataset('state_labels', data=self.assignments_file['state_labels'][...])

        # Put nice labels on things
        for ds in (self.output_file, durations_count_ds, cond_fluxes_ds, total_fluxes_ds):
            h5io.stamp_iter_range(ds, start_iter, stop_iter)

        # Calculate instantaneous rate matrices and trace trajectories
        last_state = None
        pi.new_operation('Tracing trajectories', iter_count)
        for iiter, n_iter in enumerate(range(start_iter, stop_iter)):
            # Get data from the main HDF5 file
            iter_group = self.data_reader.get_iter_group(n_iter)
            seg_index = iter_group['seg_index']
            nsegs, npts = iter_group['pcoord'].shape[0:2]
            weights = seg_index['weight']
            # parent_ids = seg_index['parent_id']
            parent_ids = self.data_reader.parent_id_dsspec.get_iter_data(n_iter)

            # Get bin and traj. ensemble assignments from the previously-generated assignments file
            assignment_iiter = h5io.get_iteration_entry(self.assignments_file, n_iter)
            bin_assignments = np.require(
                self.assignments_file['assignments'][assignment_iiter + np.s_[:nsegs, :npts]], dtype=index_dtype
            )
            label_assignments = np.require(
                self.assignments_file['trajlabels'][assignment_iiter + np.s_[:nsegs, :npts]], dtype=index_dtype
            )
            state_assignments = np.require(
                self.assignments_file['statelabels'][assignment_iiter + np.s_[:nsegs, :npts]], dtype=index_dtype
            )

            # Prepare to run analysis
            cond_fluxes = np.zeros((nstates, nstates), weight_dtype)
            total_fluxes = np.zeros((nstates,), weight_dtype)
            cond_counts = np.zeros((nstates, nstates), np.uint)
            total_counts = np.zeros((nstates,), np.uint)
            durations = []

            # Estimate macrostate fluxes and calculate event durations using trajectory tracing
            # state is opaque to the find_macrostate_transitions function
            dt = 1.0 if npts == 1 else 1.0 / (npts - 1)
            state = _fast_transition_state_copy(iiter, nstates, parent_ids, last_state)
            find_macrostate_transitions(
                nstates,
                weights,
                label_assignments,
                state_assignments,
                dt,
                state,
                cond_fluxes,
                cond_counts,
                total_fluxes,
                total_counts,
                durations,
            )
            last_state = state

            # Store trace-based kinetics data
            cond_fluxes_ds[iiter] = cond_fluxes
            total_fluxes_ds[iiter] = total_fluxes
            arrival_counts_ds[iiter] = total_counts
            cond_arrival_counts_ds[iiter] = cond_counts

            durations_count_ds[iiter] = len(durations)
            if len(durations) > 0:
                durations_ds.resize((iter_count, max(len(durations), durations_ds.shape[1])))
                durations_ds[iiter, : len(durations)] = durations

            # Do a little manual clean-up to prevent memory explosion
            del iter_group, weights, parent_ids, bin_assignments, label_assignments, state, cond_fluxes, total_fluxes
            pi.progress += 1




# A cythoned version of the original function of the stats_process function,
# based on _kinetics.pyx

from __future__ import print_function,division
import cython
import numpy
import numpy as np
import h5py
from scipy.sparse import csgraph
import warnings
from collections import Counter
cimport numpy
cimport numpy as np
cimport scipy.linalg.cython_lapack as cl
cimport scipy.linalg
import scipy.linalg
from libc.math cimport isnan

ctypedef numpy.uint16_t index_t
ctypedef numpy.float64_t weight_t
ctypedef numpy.uint8_t bool_t
ctypedef numpy.int64_t trans_t
ctypedef numpy.uint_t uint_t # 32 bits on 32-bit systems, 64 bits on 64-bit systems
ctypedef unsigned short Ushort
ctypedef double complex Cdouble

weight_dtype = numpy.float64
index_dtype = numpy.uint16
bool_dtype = numpy.bool_
intc_dtype = numpy.intc

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef stats_process(numpy.ndarray[index_t, ndim=2] bin_assignments,
                    numpy.ndarray[weight_t, ndim=1] weights,
                    numpy.ndarray[weight_t, ndim=2] fluxes,
                    numpy.ndarray[weight_t, ndim=1] populations,
                    numpy.ndarray[trans_t, ndim=2] trans,
                    numpy.ndarray[index_t, ndim=2] mask,
                    str interval='timepoint'                        ):
    cdef:
        Py_ssize_t i,k
        index_t ibin,fbin,nsegs,npts
    nsegs = bin_assignments.shape[0]
    npts = bin_assignments.shape[1]

    if interval == 'timepoint':
        for i in xrange(0,npts - 1):
            for k in xrange(nsegs):
                ibin = bin_assignments[k,i]
                fbin = bin_assignments[k, i + 1]

                if mask[k, 0] == 1:
                    continue

                w = weights[k]

                fluxes[ibin, fbin] += w
                trans[ibin, fbin] += 1
                populations[ibin] += w
        return

    if interval == 'iteration':
        for k in xrange(nsegs):
            # Should this be 0?
            # .... this should super be 0.  What?
            ibin = bin_assignments[k,0]
            fbin = bin_assignments[k, npts - 1]

            if mask[k, 0] == 1:
                continue

            w = weights[k]

            fluxes[ibin, fbin] += w
            trans[ibin, fbin] += 1
            populations[ibin] += w
        return

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int normalize(weight_t[:,:] m, Py_ssize_t nfbins) nogil:

    cdef:
        weight_t row_sum
        Py_ssize_t x, y

    for y in range(nfbins):
        row_sum = 0
        for x in range(nfbins):
            row_sum += m[y,x]
        if row_sum != 0:
            for x in range(nfbins):
                m[y,x] /= row_sum
    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef reweight_for_c(rows, cols, obs, flux, insert, indices, nstates, nbins, state_labels, state_map, nfbins, istate, jstate, stride, bin_last_state_map, bin_state_map, return_obs, obs_threshold=1):



    # Instead of pulling in start and stop, we'll pull in a list of indices.
    # This way, it should support the bootstrap.
    cdef:
        int[:] _rows, _cols, _obs, _ins, _nrows, _ncols, _nobs, _visited
        long[:]  _bin_last_state_map
        weight_t[:] _flux, _total_pop, _rw_bin_probs, _nflux, _rw_color_probs, _rw_state_probs
        double[:] _eigvals, _eigvalsi
        int n_trans, _nstates, lind, _nfbins, _stride, _obs_threshold, nnz, nlind, i, j, _istate, _jstate
        Ushort[:] _indices, _bin_state_map
        weight_t _return_value

        Ushort[:] _new_indices

        weight_t[:,:] _total_fluxes, _transition_matrix, _rw_state_flux, _strong_transition_matrix
        double[:,:] _WORK, _eigvecs
        int[:,:] _total_obs, _graph
        #bint _return_flux, _return_states, _return_color
        str _return_obs
        #double[:] eigvals, eigvalsi
        #double[:,:] eigvecs, WORK



    # CREATE NUMPY ARRAYS
    # This is a temporary measure that fixes some segfaults, which implies I'm probably off by
    # a little bit.  Memory heavy, but whatever.
    # It breaks depending on things, so I need to root that out.  Clearly, nnz is larger than that.
    _flux = flux
    nnz = len(flux)
    lind = indices.shape[0]
    nlind = indices.shape[0]*stride
    total_fluxes = np.zeros((nfbins, nfbins), weight_dtype)
    total_obs = np.zeros((nfbins, nfbins), intc_dtype)
    transition_matrix = np.zeros((nfbins, nfbins), weight_dtype)
    strong_transition_matrix = np.zeros((nfbins, nfbins), weight_dtype)
    visited = np.zeros((nfbins), intc_dtype)
    graph = np.zeros((nfbins, nfbins+1), dtype=intc_dtype)
    rw_bin_probs = np.zeros(nfbins, weight_dtype)
    new_indices = np.zeros(((nlind)), dtype=indices.dtype)
    rw_state_flux = np.zeros((nstates, nstates), np.float64)
    state_flux = np.zeros((nstates, nstates), weight_dtype)
    eigvals = np.zeros((nfbins), np.float64)
    eigvalsi = np.zeros((nfbins), np.float64)
    eigvecs = np.zeros((nfbins, nfbins), np.float64)
    WORK = np.zeros((nfbins*4, nfbins*4), np.float64)
    rw_color_probs = np.zeros((nstates), weight_dtype)
    rw_state_probs = np.zeros((nbins), weight_dtype)


    # CREATE MEMORYVIEWS
    # These are for what we sent in...
    _rows = rows
    _cols = cols
    _obs = obs
    _flux = flux
    _ins = insert
    _nstates = nstates
    # ... these are for functions we'll be using.
    _total_fluxes = total_fluxes
    _total_obs = total_obs
    _transition_matrix = transition_matrix
    _strong_transition_matrix = strong_transition_matrix
    _nfbins = nfbins
    _stride = stride
    _obs_threshold = obs_threshold
    _indices = indices
    _new_indices = new_indices
    _rw_state_flux = rw_state_flux
    _rw_bin_probs = rw_bin_probs
    _eigvals = eigvals
    _eigvalsi = eigvalsi
    _eigvals = eigvals
    _eigvecs = eigvecs
    _WORK = WORK
    _graph = graph
    _visited = visited
    _rw_color_probs = rw_color_probs
    _rw_state_probs = rw_state_probs
    _bin_last_state_map = bin_last_state_map
    _bin_state_map = bin_state_map
    _istate = istate
    _jstate = jstate
    _return_obs = return_obs


    #NOGIL
    # Reconstruct dataset.  We're just passing the same thing back and forth between functions.
    with nogil:
        for i in range(_nfbins):
            for j in range(1, _nfbins+1):
                _graph[i, j] = _nfbins
        regenerate_subsampled_indices(_indices, _new_indices, lind, _stride)
        accumulate_fluxes(_rows, _cols, _obs, _flux, _ins, _new_indices, nnz, _transition_matrix, nlind)
        accumulate_obs(_rows, _cols, _obs, _flux, _ins, _new_indices, nnz, _total_obs, nlind)

        remove_under_obs(_transition_matrix, _total_obs, _obs_threshold, _nfbins)
        normalize(_transition_matrix, _nfbins)
        steadystate_solve(_transition_matrix, _strong_transition_matrix, _rw_bin_probs, _nfbins, _eigvals, _eigvalsi, _eigvecs, _WORK, _graph, _visited)

        for i in range(_nfbins):
            _rw_color_probs[_bin_last_state_map[i]] += _rw_bin_probs[i]
            _rw_state_probs[_bin_state_map[i]] += _rw_bin_probs[i]


        calc_state_flux(_transition_matrix, _rw_bin_probs, _bin_last_state_map, _bin_state_map, _nstates, _rw_state_flux, _nfbins)

    # This allows us to use the same function for all three types.
    # Return conditional fluxes.
    if _return_obs == b'F':
        _return_value = _rw_state_flux[_istate,_jstate]
        if isnan(_return_value) is True:
            return 0.0
        else:
            return _return_value
    # Return state probabilities.
    elif _return_obs == b'S':
        _return_value = _rw_state_probs[_istate]
        if isnan(_return_value) is True:
            return 0.0
        else:
            return _return_value
    # Return color (ensemble) probabilities
    elif _return_obs == b'C':
        _return_value = _rw_color_probs[_istate]
        if isnan(_return_value) is True:
            return 0.0
        else:
            return _return_value
    # Return the rates.
    elif _return_obs == b'R':
        if _rw_color_probs[_istate] != 0.0:
            _return_value = (_rw_state_flux[_istate,_jstate] / (_rw_color_probs[_istate] / (_rw_color_probs[_istate] + _rw_color_probs[_jstate])))
            if isnan(_return_value) is True:
                return 0.0
            else:
                return _return_value
        else:
            # We have no ensemble probability, and as such, cannot have a flux.
            return 0.0
    # Return the populations.
    elif _return_obs == b'P':
        return rw_bin_probs

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int regenerate_subsampled_indices(Ushort[:] iin, Ushort[:] iout, int ilen, int stride) nogil:

    cdef:
        int i, si

    # go over the range of all indices within iin
    for i in range(ilen):
        # Run over the length of the stride.
        for si in range(stride):
            iout[(i*stride)+si] = iin[i] + si

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int accumulate_fluxes(int[:] hrows, int[:] hcols, int[:] hobs, weight_t[:] hflux, int[:] hins, Ushort[:] iterations, Py_ssize_t nnz, weight_t[:,:] total_fluxes, int itermax) nogil:

    cdef:
        index_t curriter, elem, iiter, ipop
        long ilem

    curriter = 0

    for iter in range(itermax):
        iiter = iterations[iter]
        for ilem in range(hins[iiter], hins[iiter+1]):
            # Not sure if this is necessary, here...
            if ilem < nnz and iiter+1 < itermax:
                total_fluxes[hrows[ilem], hcols[ilem]] += hflux[ilem]

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int accumulate_obs(int[:] hrows, int[:] hcols, int[:] hobs, weight_t[:] hflux, int[:] hins, Ushort[:] iterations, Py_ssize_t nnz, int[:,:] total_obs, int itermax) nogil:

    cdef:
        index_t curriter, elem, iiter, ipop
        long ilem

    curriter = 0

    for iter in range(itermax):
        iiter = iterations[iter]
        for ilem in range(hins[iiter], hins[iiter+1]):
            if ilem < nnz and iiter+1 < itermax:
                total_obs[hrows[ilem], hcols[ilem]] += hobs[ilem]

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int remove_under_obs(weight_t[:,:] flux, int[:,:] obs, int threshold, int nbins) nogil:

    cdef:
        int x, y

    for x in range(nbins):
        for y in range(nbins):
            if obs[x,y] < threshold:
                flux[x,y] = 0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int calc_state_flux(weight_t[:, :] trans_matrix, weight_t[:] bin_probs, long[:] bin_last_state_map, Ushort[:] bin_state_map, int nstates, weight_t[:,:] state_flux, int K_shape) nogil:

    cdef:
        int i, j, ii, jj


    for i in range(K_shape):
        for j in range(K_shape):

            ii = bin_last_state_map[i]
            jj = bin_state_map[j]

            if jj != nstates:
                state_flux[ii, jj] += (trans_matrix[i, j] * bin_probs[i])

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int steadystate_solve(weight_t[:,:] K, weight_t[:,:] K_mod, weight_t[:] bin_prob, int K_shape, double[:] eigvals, double[:] eigvalsi, double[:,:] eigvecs, double[:,:] WORK, int[:,:] graph, int[:] visited) nogil:

    cdef:
        int[:] components
        int[:,:] _graph
        double max, eigsum
        int n_components, components_assignments, largest_component, maxi, x, y, n, INFO, LWORK, i, j

        # POINTERS

        int  *_INFO, *_K_shape, *_LWORK
        double *_K_mod, *_eigvals, *_eigvecs, *_WORK, *_eigvalsi
    _K_shape = &K_shape
    _INFO = &INFO
    _LWORK = &LWORK
    INFO = 0
    LWORK = K_shape * 4
    _K_mod = &K_mod[0,0]
    _eigvals = &eigvals[0]
    _eigvecs = &eigvecs[0,0]
    _eigvalsi = &eigvalsi[0]
    _WORK = &WORK[0,0]
    _graph = graph

    for i in range(K_shape):
        if visited[i] == 0:
            visited[i] = 1
            return_strong_component(K, K_shape, _graph, i, i, visited)
    n = 0
    for i in range(K_shape):
        if graph[i, 0] >= graph[n, 0]:
            n = i
    # I suspect this may be giving us issues?
    #components = _graph[n, :K_shape+1]

    maxi = 0
    eigsum = 0.0
    # This all works!
    for x in range(K_shape):
        #i = components[x+1]
        i = graph[n, x+1]
        for y in range(K_shape):
            #j = components[y+1]
            j = graph[n, y+1]
            if i != K_shape and j != K_shape:
                K_mod[i, j] = K[i, j]
    normalize(K_mod, K_shape)
    cl.dgeev('N', 'V', _K_shape, _K_mod, _K_shape, _eigvals, _eigvalsi, _eigvecs, _K_shape, _eigvecs, _K_shape, _WORK, _LWORK, _INFO)
    for x in range(K_shape):
        if x == 0:
            max = eigvals[0]
            maxi = x
        else:
            if max < eigvals[x]:
                max = eigvals[x]
                maxi = x
    # We need to go over the whole range and pick out non K_shape elements.
    # This probably no longer needs to be done, now...
    for i in range(K_shape):
        #x = components[i+1]
        x = graph[n, i+1]
        if x != K_shape:
            #eigsum += eigvecs[maxi, components[i+1]]
            eigsum += eigvecs[maxi, x]
    for i in range(K_shape):
        #x = components[i+1]
        x = graph[n, i+1]
        if x != K_shape:
            #bin_prob[components[i+1]] = eigvecs[maxi, components[i+1]]
            #bin_prob[components[i+1]] /= eigsum
            bin_prob[x] = eigvecs[maxi, x]
            bin_prob[x] /= eigsum

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef int return_strong_component(weight_t[:,:] K, int K_shape, int[:, :] graph, int i, int z, int[:] visited) nogil:

    cdef:
        int j, y

    if graph[z, 0] == 0:
        graph[z, 0] += 1
        graph[z, 1] = i
    for j in xrange(K_shape):
        if i != j:
            if K[i, j] > 0.0:
                # Strongly connected!
                if visited[j] == 0:
                    graph[z, 0] += 1
                    y = graph[z, 0]
                    graph[z, y] = j
                    # We only want to call it when we haven't visited it before.
                    # We don't want to call, THEN modify and check.  Otherwise, we could be doing many calls.
                    visited[j] = 1
                    return_strong_component(K, K_shape, graph, j, z, visited)


    return 0



from itertools import count
import logging
import math
import pickle
from pickle import PickleError
import sys

import numpy as np

import westpa
import westpa.core.binning
from westpa.core.binning import RectilinearBinMapper
from westpa.core.data_manager import weight_dtype
from westpa.core.extloader import get_object

from .core import WESTToolComponent


EPS = np.finfo(weight_dtype).eps

log = logging.getLogger(__name__)


def mapper_from_expr(expr):
    namespace = {'numpy': np, 'np': np, 'inf': float('inf')}
    try:
        mapper = RectilinearBinMapper(eval(expr, namespace))
    except TypeError as e:
        if 'has no len' in str(e):
            raise ValueError('invalid bin boundary specification (a list of lists is required)')
        else:
            raise
    else:
        log.debug('loaded {!r} from expression {!r}'.format(mapper, expr))
        return mapper


def mapper_from_system():
    system = westpa.rc.get_system_driver()
    log.debug('loaded {!r} from {!r}'.format(system.bin_mapper, system))
    return system.bin_mapper


def mapper_from_function(funcspec):
    '''Return a mapper constructed by calling a function in a named module.
    ``funcspec`` should be formatted as ``[PATH]:MODULE.FUNC``. This function
    loads MODULE, optionally adding PATH to the search path, then returns MODULE.FUNC()'''
    if ':' in funcspec:
        (pathpart, funcpart) = funcspec.rsplit(':')
        pathinfo = ['.'] + pathpart.split(':')
    else:
        funcpart = funcspec
        pathinfo = ['.']

    fn = get_object(funcpart, ['.'] + pathinfo)
    mapper = fn()
    log.debug('loaded {!r} from {!r}'.format(mapper, fn))
    return mapper


def mapper_from_hdf5(topol_group, hashval):
    '''Retrieve the mapper identified by ``hashval`` from the given bin topology group
    ``topol_group``. Returns ``(mapper, pickle, hashval)``'''
    try:
        index_ds = topol_group['index']
        pickle_ds = topol_group['pickles']
    except KeyError:
        log.debug('binhash {} not found'.format(hashval), exc_info=True)
        raise KeyError('binhash {} not found'.format(hashval))

    n_entries = len(index_ds)
    if n_entries == 0:
        raise KeyError('hash {} not found'.format(hashval))

    chunksize = 256
    for istart in range(0, n_entries, chunksize):
        chunk = index_ds[istart : min(istart + chunksize, n_entries)]
        for i in range(len(chunk)):
            if chunk[i]['hash'] == bytes(hashval, 'utf-8'):
                pkldat = bytes(pickle_ds[istart + i, 0 : chunk[i]['pickle_len']].data)
                # mapper = pickle.loads(pkldat, encoding='latin1')
                mapper = pickle.loads(pkldat)
                log.debug('loaded {!r} from {!r}'.format(mapper, topol_group))
                log.debug('hash value {!r}'.format(hashval))
                return mapper, pkldat, hashval

    raise KeyError('hash {} not found'.format(hashval))


def mapper_from_yaml(yamlfilename):
    import yaml

    ydict = yaml.load(open(yamlfilename, 'rt'), Loader=yaml.Loader)
    ybins = ydict['bins']
    from westpa.core._rc import bins_from_yaml_dict

    # return mapper_from_dict(ybins)
    return bins_from_yaml_dict(ybins)


# We want this function to live on...
def mapper_from_dict(ybins):
    typename = ybins.pop('type')
    kwargs = ybins

    try:
        mapper_type = getattr(sys.modules['westpa.core.binning'], typename)
    except AttributeError:
        raise KeyError('unknown bin mapper type {!r}'.format(typename))

    if typename == 'RectilinearBinMapper':
        boundary_lists = kwargs.pop('boundaries')
        for ilist, boundaries in enumerate(boundary_lists):
            boundary_lists[ilist] = list(
                map((lambda x: float('inf') if (x if isinstance(x, str) else '').lower() == 'inf' else x), boundaries)
            )
        return mapper_type(boundary_lists)
    else:
        try:
            return mapper_type(**kwargs)
        except Exception:
            log.exception('exception instantiating mapper')
            raise


def write_bin_info(mapper, assignments, weights, n_target_states, outfile=sys.stdout, detailed=False):
    '''Write information about binning to ``outfile``, given a mapper (``mapper``) and the weights
    (``weights``) and bin assignments (``assignments``) of a set of segments, along with a target state
    count (``n_target_states``). If ``detailed`` is true, then per-bin information is written as well as
    summary information about all bins.'''

    norm = weights.sum()
    enorm = norm - 1.0
    enormeps = abs(enorm / EPS)
    bincounts = np.bincount(assignments, minlength=len(assignments))
    n_occupied = np.count_nonzero(bincounts)
    binweights = np.bincount(assignments, weights, minlength=len(assignments))
    nonzero_counts = bincounts > 0
    n_active = mapper.nbins - n_target_states
    weights_by_bin = [[] for _i in range(mapper.nbins)]

    min_bin_weight = binweights[nonzero_counts].min()
    max_bin_weight = binweights.max()
    min_seg_weight = weights.min()
    max_seg_weight = weights.max()

    ndec = int(math.ceil(-math.log10(1 / n_active)))

    outfile.write('{:d} segments\n'.format(len(weights)))
    outfile.write(
        '{:d} bins total, {:d} targets, {:d} ({:.{ndec}%}) occupied\n'.format(
            mapper.nbins, n_target_states, n_occupied, n_occupied / n_active, ndec=ndec
        )
    )
    outfile.write('Minimum probability by bin:     {:23.17e}\n'.format(min_bin_weight))
    outfile.write('Maximum probability by bin:     {:23.17e}\n'.format(max_bin_weight))
    outfile.write('Dynamic range (by bin):         {:g} kT\n'.format(-math.log(min_bin_weight / max_bin_weight)))
    outfile.write('Minimum probability by segment: {:23.17e}\n'.format(min_seg_weight))
    outfile.write('Maximum probability by segment: {:23.17e}\n'.format(max_seg_weight))
    outfile.write('Dynamic range (by segment):     {:g} kT\n'.format(-math.log(min_seg_weight / max_seg_weight)))
    outfile.write('Norm = {:g}, error in norm = {:g} ({:.0f} epsilon)\n'.format(norm, enorm, enormeps))

    if not detailed:
        return

    outfile.write('\n')
    for iseg, weight in enumerate(weights):
        weights_by_bin[assignments[iseg]].append(weight)

    mw = max(6, max(len(str(mapper.nbins - 1)), len(str(bincounts.max()))))
    fmt = '    '.join(
        [
            '{ibin:{mw}d}',
            '{nwalkers:{mw}d}',
            '{weight:23.17e}',
            '{min_weight:23.17e}',
            '{max_weight:23.17e}',
            '{weight_ratio:12.5f}',
            '{label}\n',
        ]
    )
    outfile.write(
        '{:>{mw}s}    {:>{mw}s}    {:23s}    {:23s}    {:23s}    {:12s}    {}\n'.format(
            'Index', 'Count', 'Total weight', 'Min seg weight', 'Max seg weight', 'Weight ratio', 'Label', mw=mw
        )
    )

    for ibin, label, nwalkers, total_weight, seg_weights in zip(count(), mapper.labels, bincounts, binweights, weights_by_bin):
        if nwalkers > 0:
            min_seg_weight = min(seg_weights)
            max_seg_weight = max(seg_weights)
            weight_ratio = max_seg_weight / min_seg_weight
        else:
            min_seg_weight = 0
            max_seg_weight = 0
            weight_ratio = 0
        outfile.write(
            fmt.format(
                ibin=ibin,
                label=label,
                nwalkers=nwalkers,
                weight=total_weight,
                min_weight=min_seg_weight,
                max_weight=max_seg_weight,
                weight_ratio=weight_ratio,
                mw=mw,
            )
        )


def write_bin_labels(mapper, dest, header='# bin labels:\n', fmt='# bin {index:{max_iwidth}d} -- {label!s}\n'):

    '''Print labels for all bins in ``mapper`` to the file-like object``dest``.

    If provided, ``header`` is printed prior to any labels. A number of expansions
    are available in ``header``:
      * ``mapper`` -- the mapper itself (from which most of the following can be obtained)
      * ``classname``  -- the class name of the mapper
      * ``nbins`` -- number of bins in the mapper

    The ``fmt`` string specifies how bin labels are to be printed. A number of
    expansions are available in ``fmt``:
      * ``index`` -- the zero-based index of the bin
      * ``label`` -- the label of the bin
      * ``max_iwidth`` -- the maximum width (in characters) of the bin index, for pretty alignment
    '''

    if header:
        dest.write(header.format(mapper=mapper, classname=mapper.__class__.__name__, nbins=mapper.bins))

    max_iwidth = len(str(mapper.nbins - 1))
    for (ibin, label) in enumerate(mapper.labels):
        dest.write(fmt.format(index=ibin, label=label, max_iwidth=max_iwidth))


class BinMappingComponent(WESTToolComponent):
    '''Component for obtaining a bin mapper from one of several places based on
    command-line arguments. Such locations include an HDF5 file that contains
    pickled mappers (including the primary WEST HDF5 file), the system object,
    an external function, or (in the common case of rectilinear bins) a
    list of lists of bin boundaries.

    Some configuration is necessary prior to calling process_args() if loading a
    mapper from HDF5. Specifically, either set_we_h5file_info() or
    set_other_h5file_info() must be called to describe where to find the
    appropriate mapper. In the case of set_we_h5file_info(), the mapper used for
    WE at the end of a given iteration will be loaded. In the case of
    set_other_h5file_info(), an arbitrary group and hash value are specified;
    the mapper corresponding to that hash in the given group will be returned.

    In the absence of arguments, the mapper contained in an existing HDF5 file
    is preferred; if that is not available, the mapper from the system driver
    is used.

    This component adds the following arguments to argument parsers:

      --bins-from-system
          Obtain bins from the system driver

      --bins-from-expr=EXPR
          Construct rectilinear bins by parsing EXPR and calling
          RectilinearBinMapper() with the result. EXPR must therefore be a list of
          lists.

      --bins-from-function=[PATH:]MODULE.FUNC
          Call an external function FUNC in module MODULE (optionally adding PATH
          to the search path when loading MODULE) which, when called, returns a
          fully-constructed bin mapper.

      --bins-from-file
          Load bin definitions from a YAML configuration file.

      --bins-from-h5file
          Load bins from the file being considered; this is intended to mean the
          master WEST HDF5 file or results of other binning calculations, as
          appropriate.
    '''

    def __init__(self):

        # The final mapper
        self.mapper = None
        self.mapper_hash = None
        self.mapper_pickle = None
        self.mapper_source = None
        self.mapper_source_desc = None
        self.bin_target_counts = None
        self.target_counts_required = False

        # Cues for where to load a mapper from, for the external file case
        self.mapper_source_group = None
        self.mapper_source_n_iter = None
        self.mapper_source_hash = None

        self._parse_target_count_args = False

    def add_args(self, parser, description='binning options', suppress=[]):
        suppressed_options = set(suppress)

        group = parser.add_argument_group(description)
        egroup = group.add_mutually_exclusive_group()
        if '--bins-from-system' not in suppressed_options:
            egroup.add_argument(
                '--bins-from-system',
                action='store_true',
                help='''Bins are constructed by the system driver specified in the WEST configuration file
                                (default where stored bin definitions not available).''',
            )
        if '--bins-from-expr' not in suppressed_options:
            egroup.add_argument(
                '--bins-from-expr',
                '--binbounds',
                dest='bins_from_expr',
                help='''Construct bins on a rectilinear grid according to the given BINEXPR. This must
                                be a list of lists of bin boundaries (one list of bin boundaries for each dimension
                                of the progress coordinate), formatted as a Python expression. E.g. "[[0,1,2,4,inf],[-inf,0,inf]]".
                                The numpy module and the special symbol "inf" (for floating-point infinity) are available
                                for use within BINEXPR.''',
            )
        if '--bins-from-function' not in suppressed_options:
            egroup.add_argument(
                '--bins-from-function',
                '--binfunc',
                dest='bins_from_function',
                help='''Supply an external function which, when called, returns a properly constructed
                                bin mapper which will then be used for bin assignments. This should be formatted as
                                "[PATH:]MODULE.FUNC", where the function FUNC in module MODULE will be used; the optional
                                PATH will be prepended to the module search path when loading MODULE.''',
            )
        if '--bins-from-file' not in suppressed_options:
            egroup.add_argument(
                '--bins-from-file',
                '--binfile',
                dest='bins_from_file',
                metavar='BINFILE',
                help='''Load bin specification from the YAML file BINFILE. This currently
                                takes the form {'bins': {'type': 'RectilinearBinMapper', 'boundaries':
                                [[boundset1], [boundset2], ... ]}}; only rectilinear bin bounds are supported.''',
            )

        if '--bins-from-h5file' not in suppressed_options:
            egroup.add_argument(
                '--bins-from-h5file',
                action='store_true',
                help='''Load bin specification from the data file being examined
                                (default where stored bin definitions available).''',
            )

    def add_target_count_args(self, parser, description='bin target count options'):
        '''Add options to the given parser corresponding to target counts.'''

        self._parse_target_count_args = True
        group = parser.add_argument_group(description)
        egroup = group.add_mutually_exclusive_group()

        egroup.add_argument(
            '--target-counts',
            help='''Use TARGET_COUNTS instead of stored or system driver target counts.
                            TARGET_COUNTS is a comma-separated list of integers. As a special case, a single
                            integer is acceptable, in which case the same target count is used for all bins.''',
        )
        egroup.add_argument(
            '--target-counts-from',
            metavar='FILENAME',
            help='''Read target counts from the text file FILENAME instead of using stored or system
                            driver target counts. FILENAME must contain a list of integers, separated by arbitrary
                            whitespace (including newlines).''',
        )

    def process_args(self, args):

        # User may have suppressed any of these arguments
        bins_from_system = getattr(args, 'bins_from_system', None)
        bins_from_expr = getattr(args, 'bins_from_expr', None)
        bins_from_function = getattr(args, 'bins_from_function', None)
        bins_from_file = getattr(args, 'bins_from_file', None)
        bins_from_h5file = getattr(args, 'bins_from_h5file', None)

        if not any([bins_from_system, bins_from_expr, bins_from_function, bins_from_h5file, bins_from_file]):
            log.debug('no argument provided')
            if self.mapper_source_group is None:
                log.info('using bins from system file')
                bins_from_system = True
            else:
                log.info('using bins from HDF5 file')
                bins_from_h5file = True

        if bins_from_h5file:
            # it is an error to call process_args() prior to setting source group and hash,
            # preferably using the set_*_h5file_info() functions.
            assert self.mapper_source_group is not None
            assert self.mapper_source_hash is not None
            self.mapper, self.mapper_pickle, self.mapper_hash = mapper_from_hdf5(self.mapper_source_group, self.mapper_source_hash)
            self.mapper_source = 'file'
            self.mapper_source_desc = 'HDF5 file {} (group "{}")'.format(
                self.mapper_source_group.file.filename, self.mapper_source_group.name
            )
        elif bins_from_file:
            self.mapper = mapper_from_yaml(args.bins_from_file)
            self.mapper_source = args.bins_from_file
            self.mapper_source_desc = 'YAML file {!r}'.format(args.bins_from_file)
        elif bins_from_system:
            self.mapper = mapper_from_system()
            self.mapper_source = 'system'
            self.mapper_source_desc = 'system driver'
        elif bins_from_expr:
            self.mapper = mapper_from_expr(args.bins_from_expr)
            self.mapper_source = 'expr'
            self.mapper_source_desc = 'boundary expression'
        elif bins_from_function:
            self.mapper = mapper_from_function(args.bins_from_function)
            self.mapper_source = 'func'
            self.mapper_source_desc = 'external function'

        if self.mapper and not self.mapper_hash:
            try:
                self.mapper_pickle, self.mapper_hash = self.mapper.pickle_and_hash()
            except PickleError as e:
                log.debug('could not pickle bin mapper: {}'.format(e))
                self.mapper_pickle = self.mapper_hash = None

        log.info('loaded mapper {!r} from {}'.format(self.mapper, self.mapper_source_desc))

        if self._parse_target_count_args and self.target_counts_required:
            import re

            if args.target_counts is not None:
                self.bin_target_counts = np.array(list(map(int, re.split(r'\s*,\s*', args.target_counts))))
            elif args.target_counts_from is not None:
                self.bin_target_counts = np.array(list(map(int, re.split(r'\s*,\s*', open(args.target_counts_from, 'rt').read()))))
            else:
                # if target counts required, they will have already been loaded from a master
                if self.bin_target_counts is None and self.target_counts_required:
                    raise EnvironmentError('target counts are required but none have been provided')

            if len(self.bin_target_counts) == 1:
                flat_target_count = self.bin_target_counts
                self.bin_target_counts = np.empty((self.mapper.nbins,), np.int)
                self.bin_target_counts[:] = flat_target_count
            log.debug('bin target counts = {!r}'.format(self.bin_target_counts))

    def set_we_h5file_info(self, n_iter=None, data_manager=None, required=False):
        '''Set up to load a bin mapper from the master WEST HDF5 file. The mapper is actually loaded
        from the file when self.load_bin_mapper() is called, if and only if command line arguments
        direct this. If ``required`` is true, then a mapper must be available at iteration ``n_iter``,
        or else an exception will be raised.'''

        data_manager = data_manager or westpa.rc.get_data_manager()
        n_iter = n_iter or data_manager.current_iteration
        iter_group = data_manager.get_iter_group(n_iter)

        try:
            self.mapper_source_group = data_manager.we_h5file['/bin_topologies']
            self.mapper_source_group['index']  # raises KeyError if missing
            self.mapper_source_group['pickles']  # raises KeyError if missing
            self.mapper_source_hash = iter_group.attrs['binhash']
        except TypeError:
            if required:
                raise EnvironmentError('HDF5 file not open')
        except KeyError:
            if required:
                raise EnvironmentError('bin topologies not available')

        try:
            self.bin_target_counts = iter_group['bin_target_counts'][...]
        except KeyError:
            pass

    def set_other_h5file_info(self, topology_group, hashval):
        '''Set up to load a bin mapper from (any) open HDF5 file, where bin topologies are
        stored in ``topology_group`` (an h5py Group object) and the desired mapper has hash
        value ``hashval``. The mapper itself is loaded when self.load_bin_mapper() is called.'''

        self.mapper_source_group = topology_group
        self.mapper_source_hash = hashval



# Copyright (C) 2013 Matthew C. Zwier and Lillian T. Chong
#
# This file is part of WESTPA.
#
# WESTPA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# WESTPA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with WESTPA.  If not, see <http://www.gnu.org/licenses/>.

import logging
import numpy as np
import pickle

log = logging.getLogger(__name__)
from westpa.tools.core import WESTTool
from westpa.core.data_manager import n_iter_dtype
from westpa.tools.progress import ProgressIndicatorComponent
from westpa.core import h5io
from westpa.tools.core import WESTMultiTool

# from westtools.dtypes import iter_block_ci_dtype as ci_dtype
import gc

# from pympler.tracker import SummaryTracker

ci_dtype = np.dtype(
    [
        ('iter_start', n_iter_dtype),
        ('iter_stop', n_iter_dtype),
        ('expected', np.float64),
        ('ci_lbound', np.float64),
        ('ci_ubound', np.float64),
        ('corr_len', n_iter_dtype),
        ('variance', np.float64),
        ('stderrormean', np.float64),
    ]
)

# directory locations are stored in a .yaml file with this format:
# ---
# PATHS: ['/path/to/simulation/1','/path/to/simulation/2',...,
# '/path/to/simulation/n']

# Straight up stolen from the data manager.  In the future, maybe I can just sort it by subbing in the appropriate values.


def get_bin_mapper(we_h5file, hashval):
    '''Look up the given hash value in the binning table, unpickling and returning the corresponding
    bin mapper if available, or raising KeyError if not.'''

    # Convert to a hex digest if we need to
    try:
        hashval = hashval.hexdigest()
    except AttributeError:
        pass

    while True:
        # these will raise KeyError if the group doesn't exist, which also means
        # that bin data is not available, so no special treatment here
        try:
            binning_group = we_h5file['/bin_topologies']
            index = binning_group['index']
            pkl = binning_group['pickles']
        except KeyError:
            raise KeyError('hash {} not found. Could not retrieve binning group'.format(hashval))

        n_entries = len(index)
        if n_entries == 0:
            raise KeyError('hash {} not found. No entries in index'.format(hashval))

        chunksize = 1024

        for istart in range(0, n_entries, chunksize):
            chunk = index[istart : min(istart + chunksize, n_entries)]
            for i in range(len(chunk)):
                if chunk[i]['hash'] == hashval:
                    pkldat = bytes(pkl[istart + i, 0 : chunk[i]['pickle_len']].data)
                    mapper = pickle.loads(pkldat)
                    log.debug('loaded {!r} from {!r}'.format(mapper, binning_group))
                    log.debug('hash value {!r}'.format(hashval))
                    return mapper

        raise KeyError('hash {} not found'.format(hashval))


class WMultiWest(WESTMultiTool):
    prog = 'w_multi_west'
    description = '''\
Tool designed to combine multiple WESTPA simulations while accounting for
reweighting.  Test code thus far.
-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''

    def __init__(self):
        super(WESTTool, self).__init__()
        self.progress = ProgressIndicatorComponent()
        # We no longer care about a lot of this.
        self.ntrials = 0
        self.nstates = 0
        self.kin_trial = {}
        self.west = {}
        self.niters = 0

    def add_args(self, parser):
        self.progress.add_args(parser)
        iogroup = parser.add_argument_group('input/output options')
        iogroup.add_argument('-o', '--output-file', default='multi.h5', help='''The name of the output file to store results in.''')
        iogroup.add_argument(
            '-W',
            '--west',
            '--WEST_H5FILE',
            default='west.h5',
            help='''The name of the main .h5 file inside each simulation
                             directory''',
        )
        iogroup.add_argument('-a', '--aux', action='append', help='''Names of additional auxiliary datasets to be combined''')
        iogroup.add_argument('-aa', '--auxall', action='store_true', help='''Combine all auxiliary datasets. Default: False''')
        iogroup.add_argument('-nr', '--no-reweight', action='store_true', help='''Do not reweight. Default: False''')

    def open_files(self):
        self.output_file = h5io.WESTPAH5File(self.output_file, 'w', creating_program=True)
        h5io.stamp_creator_data(self.output_file)

        opened_files = self.generate_file_list([self.west])
        self.westH5 = opened_files[self.west]
        # Just some temp things while I clean everything up...
        # west_files = self.westH5
        # Determine max iteration ...

        # We can't really use the old method anymore, as we need to calculate rates in the bootstrap.
        # Ergo, we're going to load things like w_kinavg, but that's all.
        # We'll just load them up and store them internally, for the moment.

    def process_args(self, args):
        self.progress.process_args(args)
        self.output_file = args.output_file
        self.output_file_name = args.output_file
        self.west = args.west
        self.sims = args.sims
        self.aux = args.aux
        self.auxall = args.auxall
        self.reweight = args.no_reweight

    def total_number_of_walkers(self):
        self.total_walkers = [0] * self.niters
        for key, west in self.westH5.items():
            # Sometimes, we're smaller or larger by one.  Hm.
            try:
                self.total_walkers[:] += west['summary'][:-1]['n_particles']
            except (ValueError):
                self.total_walkers[:] += west['summary'][:-1]['n_particles'][: len(self.total_walkers)]

    class Segment:
        def __init__(self, weight=0, iteration=0, simid=0, recycled_in=0):
            self.weight = weight
            self.iteration = iteration
            self.simid = simid
            self.recycled_in = recycled_in

    def go(self):
        pi = self.progress.indicator
        with pi:
            pi.new_operation('Initializing')
            self.open_files()
            self.total_number_of_walkers()
            if self.auxall is True:
                self.aux = list(self.westH5[1]['iterations/iter_00000001/auxdata'].keys())
            # Create a giant WEST.h5 file, separating the individual walkers, and renormalizing the weights.
            # It should then be compatible with existing toolsets.
            # Isn't really going to start with auxdata, but we'll add it in.

            # self.niters = 500
            # Initialize data manager...
            # Just bullshit for the current system.
            # self.niters = self.westH5[1].attrs['west_current_iteration'] - 1
            # print(self.niters, len(self.westH5))
            # self.data_manager = data_manager.WESTDataManager()
            westh5 = []
            self.source_sinks = []
            self.n_sims = {}
            for ifile, (key, west) in enumerate(self.westH5.items()):
                d = {'west': west, 'wm': None, 'rt': None, 'remove_next_cycle': [], 'seg_index': None}
                # We're getting the bin mapper, then setting the recycling target...
                binhash = west['iterations/iter_{0:08d}'.format(2)].attrs['binhash']
                bin_mapper = get_bin_mapper(west, bytes(binhash, 'utf-8'))
                try:
                    d['rt'] = bin_mapper.assign(west['tstates']['0']['pcoord'][...])[0]
                    self.source_sinks.append(bin_mapper.assign(west['tstates']['0']['pcoord'][...])[0])
                except KeyError:
                    d['rt'] = None
                    self.source_sinks.append(None)
                    pass
                # We're going to make a set of source and sink states that we can iterate through, eventually.
                # Keep a count of how many simulations for this particular recycling target we have...
                try:
                    self.n_sims[d['rt']] += 1
                except KeyError:
                    self.n_sims[d['rt']] = 1
                westh5.append(d)
                if ifile == 0:
                    self.niters = west.attrs['west_current_iteration'] - 1
                else:
                    self.niters = min(west.attrs['west_current_iteration'] - 1, self.niters)
            start_point = []
            self.source_sinks = list(set(self.source_sinks))
            # We'll need a global list of walkers to add to and take care of during the next round of simulations, as well as the current one.
            # We'll organize it by source and sink states.
            self.past_iter = {}
            self.futr_iter = {}
            self.past_rm = {}
            self.futr_rm = {}
            for i in self.source_sinks:
                self.past_iter[i] = []
                self.futr_iter[i] = []
                self.past_rm[i] = []
                self.futr_rm[i] = []
            print(pi.new_operation('Recreating...', self.niters))
            # tracker = SummaryTracker()
            # self.output_file.close()
            for iter in range(self.niters):
                # We have the following datasets in each iteration:
                # ibstates, which aren't important.
                # pcoord
                # seg_index
                # wtgraph
                # wtgraph is going to be a little more complex to handle, but not too bad.
                # aux data specified
                iter += 1
                ifile = 0
                # self.output_file = h5io.WESTPAH5File(self.output_file_name, 'w', creating_program=True)
                # Determine how many simulations to append or remove per west file.
                # self.segments = {}
                # for key,value in self.n_sims.items():
                #    self.segments[key] = int(np.floor(len(self.past_iter[key]) / value))

                # run_once = 0
                # total_current_sims = 0
                # for i in self.source_sinks:
                #    total_current_sims += len(self.past_iter[i])
                #    total_current_sims += len(self.past_rm[i])
                for ifile, west in enumerate(westh5):
                    westdict = west['west']
                    seg_index = westdict['iterations/iter_{0:08d}'.format(iter)]['seg_index'][...]
                    pcoord = westdict['iterations/iter_{0:08d}'.format(iter)]['pcoord'][...]
                    wtgraph = westdict['iterations/iter_{0:08d}'.format(iter)]['wtgraph'][...]
                    if self.aux:
                        auxdata = {}
                        for i in self.aux:
                            auxdata[str(i)] = westdict['iterations/iter_{0:08d}'.format(iter)]['auxdata'][str(i)][...]
                    if iter == 1 and ifile == 0:
                        new_dtype = np.dtype(seg_index.dtype.descr + [('group', '<i8')])
                    new_seg_index = np.zeros(seg_index.shape, dtype=new_dtype)
                    for dt, val in seg_index.dtype.fields.items():
                        new_seg_index[dt] = seg_index[dt]
                    new_seg_index['group'] = ifile
                    del seg_index
                    seg_index = new_seg_index[...]
                    del new_seg_index
                    if ifile == 0:
                        mseg = seg_index
                        mpco = pcoord
                        mwtg = wtgraph
                        if self.aux:
                            maux = {}
                            for i in self.aux:
                                maux[str(i)] = auxdata[str(i)]
                        if iter == 1:
                            summary = westdict['summary'][...]

                        start_point.append(0)
                    if ifile != 0:
                        # print(mseg.shape, seg_index.shape, ifile)
                        # print(mpco.shape, pcoord.shape, ifile)
                        # print(mwtg.shape, wtgraph.shape, ifile)
                        if iter != 1:
                            addition = prev_start_point[ifile]  # noqa: F821
                        else:
                            addition = mseg.shape[0]
                        seg_index['parent_id'][np.where(seg_index['parent_id'] >= 0)] += addition
                        seg_index['parent_id'][np.where(seg_index['parent_id'] < 0)] -= addition
                        seg_index['wtg_offset'] += mwtg.shape[0]
                        start_point.append(mseg.shape[0])
                        wtgraph += mwtg.shape[0]
                        mseg = np.concatenate((mseg, seg_index))
                        mpco = np.concatenate((mpco, pcoord))
                        mwtg = np.concatenate((mwtg, wtgraph))
                        if self.aux:
                            for i in self.aux:
                                maux[str(i)] = np.concatenate((maux[str(i)], auxdata[str(i)]))
                    ifile += 1
                    del seg_index, pcoord, wtgraph, westdict
                    if self.aux:
                        del auxdata
                gc.collect()
                # Make a real copy to use in the next iteration.
                # self.past_iter = self.futr_iter.copy()
                # self.past_rm[i] = self.futr_rm.copy()
                prev_start_point = start_point  # noqa: F841
                start_point = []
                # This is... maybe wrong, actually?  Or at least, it's not ALL that is required for normalizing things.
                # We need to weight everything by 1/N, then just normalize if that normalization was wrong.  Keep the relative weights sane.
                # ... or actually, no, that's fine, nevermind, what's wrong with me?  But we'll leave it in for now.

                if not self.reweight:
                    mseg['weight'] /= mseg['weight'].sum()

                summary['n_particles'][iter - 1] = mseg.shape[0]
                summary['norm'][iter - 1] = mseg['weight'].sum()
                summary['min_seg_prob'][iter - 1] = min(mseg['weight'])
                summary['max_seg_prob'][iter - 1] = max(mseg['weight'])

                curr_iter = self.output_file.create_group('iterations/iter_{0:08d}'.format(iter))
                curr_iter.attrs['n_iter'] = iter
                ds_rate_evol = curr_iter.create_dataset('wtgraph', data=mwtg, shuffle=True, compression=9)
                ds_rate_evol = curr_iter.create_dataset('seg_index', data=mseg, shuffle=True, compression=9)
                ds_rate_evol = curr_iter.create_dataset('pcoord', data=mpco, shuffle=True, compression=9)
                if self.aux:
                    aux_iter = self.output_file.create_group('iterations/iter_{0:08d}/auxdata'.format(iter))
                    for i in self.aux:
                        ds_rate_evol = aux_iter.create_dataset(str(i), data=maux[str(i)], shuffle=True, compression=9)
                # We need to be careful about memory, here.  We are blowing uppppp.
                # We're STILL blowing up.  Criiiiiipes.
                # self.segments = {}
                del mseg, mpco, mwtg, ds_rate_evol, curr_iter  # , self.segments
                if self.aux:
                    del maux, aux_iter
                gc.collect()
                self.output_file.flush()
                # self.output_file.close()
                # print("How big is our summary?")
                # print(sys.getsizeof(summary))
                # objgraph.show_most_common_types(limit=50)
                # objgraph.show_growth(limit=10)
                # objgraph.show_most_common_types(objects=objgraph.get_leaking_objects())
                pi.progress += 1

        pi.new_operation('Writing to file...')
        ds_rate_evol = self.output_file.create_dataset('summary', data=summary, shuffle=True, compression=9)  # noqa: F841
        self.output_file.attrs['west_current_iteration'] = self.niters
        self.output_file.attrs['west_file_format_version'] = 7
        self.output_file.attrs['west_iter_prec'] = 8
        self.output_file.attrs['westpa_fileformat_version'] = 7
        self.output_file.attrs['westpa_iter_prec'] = 8


def entry_point():
    WMultiWest().main()


if __name__ == '__main__':
    entry_point()



import logging
import math
import operator
import random

import numpy as np

import westpa
from .segment import Segment
from .states import InitialState

log = logging.getLogger(__name__)


def _group_walkers_identity(we_driver, ibin, **kwargs):
    log.debug('using we_driver._group_walkers_identity')
    bin_set = we_driver.next_iter_binning[ibin]
    list_bins = [set()]
    for i in bin_set:
        list_bins[0].add(i)
    return list_bins


class ConsistencyError(RuntimeError):
    pass


class AccuracyError(RuntimeError):
    pass


class NewWeightEntry:
    NW_SOURCE_RECYCLED = 0

    def __init__(
        self,
        source_type,
        weight,
        prev_seg_id=None,
        prev_init_pcoord=None,
        prev_final_pcoord=None,
        new_init_pcoord=None,
        target_state_id=None,
        initial_state_id=None,
    ):
        self.source_type = source_type
        self.weight = weight
        self.prev_seg_id = prev_seg_id
        self.prev_init_pcoord = np.asarray(prev_init_pcoord) if prev_init_pcoord is not None else None
        self.prev_final_pcoord = np.asarray(prev_final_pcoord) if prev_final_pcoord is not None else None
        self.new_init_pcoord = np.asarray(new_init_pcoord) if new_init_pcoord is not None else None
        self.target_state_id = target_state_id
        self.initial_state_id = initial_state_id

    def __repr__(self):
        return '<{} object at 0x{:x}: weight={self.weight:g} target_state_id={self.target_state_id} prev_final_pcoord={self.prev_final_pcoord}>'.format(
            self.__class__.__name__, id(self), self=self
        )


class WEDriver:
    '''A class implemented Huber & Kim's weighted ensemble algorithm over Segment objects.
    This class handles all binning, recycling, and preparation of new Segment objects for the
    next iteration. Binning is accomplished using system.bin_mapper, and per-bin target counts
    are from system.bin_target_counts.

    The workflow is as follows:

      1) Call `new_iteration()` every new iteration, providing any recycling targets that are
         in force and any available initial states for recycling.
      2) Call `assign()` to assign segments to bins based on their initial and end points. This
         returns the number of walkers that were recycled.
      3) Call `run_we()`, optionally providing a set of initial states that will be used to
         recycle walkers.

    Note the presence of flux_matrix, transition_matrix,
    current_iter_segments, next_iter_segments, recycling_segments,
    initial_binning, final_binning, next_iter_binning, and new_weights (to be documented soon).
    '''

    weight_split_threshold = 2.0
    weight_merge_cutoff = 1.0
    largest_allowed_weight = 1.0
    smallest_allowed_weight = 1e-323

    def __init__(self, rc=None, system=None):
        self.rc = rc or westpa.rc
        self.system = system or self.rc.get_system_driver()

        # Whether to adjust counts to exactly match target count
        self.do_adjust_counts = True

        # bin mapper and per-bin target counts (see new_iteration for initialization)
        self.bin_mapper = None
        self.bin_target_counts = None

        # Mapping of bin index to target state
        self.target_states = None

        # binning on initial points
        self.initial_binning = None

        # binning on final points (pre-WE)
        self.final_binning = None

        # binning on initial points for next iteration
        self.next_iter_binning = None

        # Flux and rate matrices for the current iteration
        self.flux_matrix = None
        self.transition_matrix = None

        # Information on new weights (e.g. from recycling) for the next iteration
        self.new_weights = None

        # Set of initial states passed to run_we() that are actually used for
        # recycling targets
        self.used_initial_states = None

        self.avail_initial_states = None

        # Make property for subgrouping function.
        self.subgroup_function = _group_walkers_identity
        self.subgroup_function_kwargs = {}

        self.process_config()
        self.check_threshold_configs()

    def process_config(self):
        config = self.rc.config

        config.require_type_if_present(['west', 'we', 'adjust_counts'], bool)

        config.require_type_if_present(['west', 'we', 'thresholds'], bool)

        self.do_adjust_counts = config.get(['west', 'we', 'adjust_counts'], True)
        log.info('Adjust counts to exactly match target_counts: {}'.format(self.do_adjust_counts))

        self.do_thresholds = config.get(['west', 'we', 'thresholds'], True)
        log.info('Obey abolute weight thresholds: {}'.format(self.do_thresholds))

        self.weight_split_threshold = config.get(['west', 'we', 'weight_split_threshold'], self.weight_split_threshold)
        log.info('Split threshold: {}'.format(self.weight_split_threshold))

        self.weight_merge_cutoff = config.get(['west', 'we', 'weight_merge_cutoff'], self.weight_merge_cutoff)
        log.info('Merge cutoff: {}'.format(self.weight_merge_cutoff))

        self.largest_allowed_weight = config.get(['west', 'we', 'largest_allowed_weight'], self.largest_allowed_weight)
        log.info('Largest allowed weight: {}'.format(self.largest_allowed_weight))

        self.smallest_allowed_weight = config.get(['west', 'we', 'smallest_allowed_weight'], self.smallest_allowed_weight)
        log.info('Smallest allowed_weight: {}'.format(self.smallest_allowed_weight))

    @property
    def next_iter_segments(self):
        '''Newly-created segments for the next iteration'''
        if self.next_iter_binning is None:
            raise RuntimeError('cannot access next iteration segments before running WE')

        for _bin in self.next_iter_binning:
            for walker in _bin:
                yield walker

    @property
    def current_iter_segments(self):
        '''Segments for the current iteration'''
        for _bin in self.final_binning:
            for walker in _bin:
                yield walker

    @property
    def next_iter_assignments(self):
        '''Bin assignments (indices) for initial points of next iteration.'''
        if self.next_iter_binning is None:
            raise RuntimeError('cannot access next iteration segments before running WE')

        for ibin, _bin in enumerate(self.next_iter_binning):
            for _walker in _bin:
                yield ibin

    @property
    def current_iter_assignments(self):
        '''Bin assignments (indices) for endpoints of current iteration.'''
        for ibin, _bin in enumerate(self.final_binning):
            for walker in _bin:
                yield ibin

    @property
    def recycling_segments(self):
        '''Segments designated for recycling'''
        if len(self.target_states):
            for (ibin, tstate) in self.target_states.items():
                for segment in self.final_binning[ibin]:
                    yield segment
        else:
            return

    @property
    def n_recycled_segs(self):
        '''Number of segments recycled this iteration'''
        count = 0
        for _segment in self.recycling_segments:
            count += 1
        return count

    @property
    def n_istates_needed(self):
        '''Number of initial states needed to support recycling for this iteration'''
        n_istates_avail = len(self.avail_initial_states)
        return max(0, self.n_recycled_segs - n_istates_avail)

    def check_threshold_configs(self):
        '''Check to see if weight thresholds parameters are valid'''
        if (not np.issubdtype(type(self.largest_allowed_weight), np.floating)) or (
            not np.issubdtype(type(self.smallest_allowed_weight), np.floating)
        ):
            try:
                # Trying to self correct
                self.largest_allowed_weight = float(self.largest_allowed_weight)
                self.smallest_allowed_weight = float(self.smallest_allowed_weight)
            except ValueError:
                # Generate error saying thresholds are invalid
                raise ValueError("Invalid weight thresholds specified. Please check your west.cfg.")

        if np.isclose(self.largest_allowed_weight, self.smallest_allowed_weight):
            raise ValueError("Weight threshold bounds cannot be identical.")
        elif self.largest_allowed_weight < self.smallest_allowed_weight:
            self.smallest_allowed_weight, self.largest_allowed_weight = self.largest_allowed_weight, self.smallest_allowed_weight
            log.warning('Swapped largest allowed weight with smallest allowed weight to fulfill inequality (largest > smallest).')

    def clear(self):
        '''Explicitly delete all Segment-related state.'''

        del self.initial_binning, self.final_binning, self.next_iter_binning
        del self.flux_matrix, self.transition_matrix
        del self.new_weights, self.used_initial_states, self.avail_initial_states

        self.initial_binning = None
        self.final_binning = None
        self.next_iter_binning = None
        self.flux_matrix = None
        self.transition_matrix = None
        self.avail_initial_states = None
        self.used_initial_states = None
        self.new_weights = None

    def new_iteration(self, initial_states=None, target_states=None, new_weights=None, bin_mapper=None, bin_target_counts=None):
        '''Prepare for a new iteration. ``initial_states`` is a sequence of all InitialState objects valid
        for use in to generating new segments for the *next* iteration (after the one being begun with the call to
        new_iteration); that is, these are states available to recycle to. Target states which generate recycling events
        are specified in ``target_states``, a sequence of TargetState objects. Both ``initial_states``
        and ``target_states`` may be empty as required.

        The optional ``new_weights`` is a sequence of NewWeightEntry objects which will
        be used to construct the initial flux matrix.

        The given ``bin_mapper`` will be used for assignment, and ``bin_target_counts`` used for splitting/merging
        target counts; each will be obtained from the system object if omitted or None.
        '''

        self.clear()

        new_weights = new_weights or []
        if initial_states is None:
            initial_states = initial_states or []

        # update mapper, in case it has changed on the system driver and has not been overridden
        if bin_mapper is not None:
            self.bin_mapper = bin_mapper
        else:
            self.bin_mapper = self.system.bin_mapper

        if bin_target_counts is not None:
            self.bin_target_counts = bin_target_counts
        else:
            self.bin_target_counts = np.array(self.system.bin_target_counts).copy()
        nbins = self.bin_mapper.nbins
        log.debug('mapper is {!r}, handling {:d} bins'.format(self.bin_mapper, nbins))

        self.initial_binning = self.bin_mapper.construct_bins()
        self.final_binning = self.bin_mapper.construct_bins()
        self.next_iter_binning = None

        flux_matrix = self.flux_matrix = np.zeros((nbins, nbins), dtype=np.float64)
        transition_matrix = self.transition_matrix = np.zeros((nbins, nbins), np.uint)

        # map target state specifications to bins
        target_states = target_states or []
        self.target_states = {}
        for tstate in target_states:
            tstate_assignment = self.bin_mapper.assign([tstate.pcoord])[0]
            self.target_states[tstate_assignment] = tstate
            log.debug('target state {!r} mapped to bin {}'.format(tstate, tstate_assignment))
            self.bin_target_counts[tstate_assignment] = 0

        # loop over recycled segments, adding entries to the flux matrix appropriately
        if new_weights:
            init_pcoords = np.empty((len(new_weights), self.system.pcoord_ndim), dtype=self.system.pcoord_dtype)
            prev_init_pcoords = np.empty((len(new_weights), self.system.pcoord_ndim), dtype=self.system.pcoord_dtype)

            for (ientry, entry) in enumerate(new_weights):
                init_pcoords[ientry] = entry.new_init_pcoord
                prev_init_pcoords[ientry] = entry.prev_init_pcoord

            init_assignments = self.bin_mapper.assign(init_pcoords)
            prev_init_assignments = self.bin_mapper.assign(prev_init_pcoords)

            for (entry, i, j) in zip(new_weights, prev_init_assignments, init_assignments):
                flux_matrix[i, j] += entry.weight
                transition_matrix[i, j] += 1

            del init_pcoords, prev_init_pcoords, init_assignments, prev_init_assignments

        self.avail_initial_states = {state.state_id: state for state in initial_states}
        self.used_initial_states = {}

    def add_initial_states(self, initial_states):
        '''Add newly-prepared initial states to the pool available for recycling.'''
        for state in initial_states:
            self.avail_initial_states[state.state_id] = state

    @property
    def all_initial_states(self):
        '''Return an iterator over all initial states (available or used)'''
        for state in self.avail_initial_states.values():
            yield state
        for state in self.used_initial_states.values():
            yield state

    def assign(self, segments, initializing=False):
        '''Assign segments to initial and final bins, and update the (internal) lists of used and available
        initial states. If ``initializing`` is True, then the "final" bin assignments will
        be identical to the initial bin assignments, a condition required for seeding a new iteration from
        pre-existing segments.'''

        # collect initial and final coordinates into one place
        all_pcoords = np.empty((2, len(segments), self.system.pcoord_ndim), dtype=self.system.pcoord_dtype)

        for iseg, segment in enumerate(segments):
            all_pcoords[0, iseg] = segment.pcoord[0, :]
            all_pcoords[1, iseg] = segment.pcoord[-1, :]

        # assign based on initial and final progress coordinates
        initial_assignments = self.bin_mapper.assign(all_pcoords[0, :, :])
        if initializing:
            final_assignments = initial_assignments
        else:
            final_assignments = self.bin_mapper.assign(all_pcoords[1, :, :])

        initial_binning = self.initial_binning
        final_binning = self.final_binning
        flux_matrix = self.flux_matrix
        transition_matrix = self.transition_matrix
        for (segment, iidx, fidx) in zip(segments, initial_assignments, final_assignments):
            initial_binning[iidx].add(segment)
            final_binning[fidx].add(segment)
            flux_matrix[iidx, fidx] += segment.weight
            transition_matrix[iidx, fidx] += 1

        n_recycled_total = self.n_recycled_segs
        n_new_states = n_recycled_total - len(self.avail_initial_states)

        log.debug(
            '{} walkers scheduled for recycling, {} initial states available'.format(
                n_recycled_total, len(self.avail_initial_states)
            )
        )

        if n_new_states > 0:
            return n_new_states
        else:
            return 0

    def _recycle_walkers(self):
        '''Recycle walkers'''

        # recall that every walker we deal with is already a new segment in the subsequent iteration,
        # so to recycle, we actually move the appropriate Segment from the target bin to the initial state bin

        self.new_weights = []

        n_recycled_walkers = len(list(self.recycling_segments))
        if not n_recycled_walkers:
            return
        elif n_recycled_walkers > len(self.avail_initial_states):
            raise ConsistencyError(
                'need {} initial states for recycling, but only {} present'.format(
                    n_recycled_walkers, len(self.avail_initial_states)
                )
            )

        used_istate_ids = set()
        istateiter = iter(self.avail_initial_states.values())
        for (ibin, target_state) in self.target_states.items():
            target_bin = self.next_iter_binning[ibin]
            for segment in set(target_bin):
                initial_state = next(istateiter)
                istate_assignment = self.bin_mapper.assign([initial_state.pcoord])[0]
                parent = self._parent_map[segment.parent_id]
                parent.endpoint_type = Segment.SEG_ENDPOINT_RECYCLED

                if log.isEnabledFor(logging.DEBUG):
                    log.debug(
                        'recycling {!r} from target state {!r} to initial state {!r}'.format(segment, target_state, initial_state)
                    )
                    log.debug('parent is {!r}'.format(parent))

                segment.parent_id = -(initial_state.state_id + 1)
                segment.pcoord[0] = initial_state.pcoord

                self.new_weights.append(
                    NewWeightEntry(
                        source_type=NewWeightEntry.NW_SOURCE_RECYCLED,
                        weight=parent.weight,
                        prev_seg_id=parent.seg_id,
                        # the .copy() is crucial, otherwise the slice of pcoords will
                        # keep the parent segments' pcoord data alive unnecessarily long
                        prev_init_pcoord=parent.pcoord[0].copy(),
                        prev_final_pcoord=parent.pcoord[-1].copy(),
                        new_init_pcoord=initial_state.pcoord.copy(),
                        target_state_id=target_state.state_id,
                        initial_state_id=initial_state.state_id,
                    )
                )

                if log.isEnabledFor(logging.DEBUG):
                    log.debug('new weight entry is {!r}'.format(self.new_weights[-1]))

                self.next_iter_binning[istate_assignment].add(segment)

                initial_state.iter_used = segment.n_iter
                log.debug('marking initial state {!r} as used'.format(initial_state))
                used_istate_ids.add(initial_state.state_id)
                target_bin.remove(segment)

            assert len(target_bin) == 0

        # Transfer newly-assigned states from "available" to "used"
        for state_id in used_istate_ids:
            self.used_initial_states[state_id] = self.avail_initial_states.pop(state_id)

    def _split_walker(self, segment, m, bin):
        '''Split the walker ``segment`` (in ``bin``) into ``m`` walkers'''
        new_segments = []
        for _inew in range(0, m):
            new_segment = Segment(
                n_iter=segment.n_iter,  # previously incremented
                weight=segment.weight / m,
                parent_id=segment.parent_id,
                wtg_parent_ids=set(segment.wtg_parent_ids),
                pcoord=segment.pcoord.copy(),
                status=Segment.SEG_STATUS_PREPARED,
            )
            new_segment.pcoord[0, :] = segment.pcoord[0, :]
            new_segments.append(new_segment)

        if log.isEnabledFor(logging.DEBUG):
            log.debug('splitting {!r} into {:d}:\n    {!r}'.format(segment, m, new_segments))

        return new_segments

    def _merge_walkers(self, segments, cumul_weight, bin):
        '''Merge the given ``segments`` in ``bin``, previously sorted by weight, into one conglomerate segment.
        ``cumul_weight`` is the cumulative sum of the weights of the ``segments``; this may be None to calculate here.'''

        if cumul_weight is None:
            cumul_weight = np.add.accumulate([segment.weight for segment in segments])

        glom = Segment(
            n_iter=segments[0].n_iter,  # assumed correct (and equal among all segments)
            weight=cumul_weight[len(segments) - 1],
            status=Segment.SEG_STATUS_PREPARED,
            pcoord=self.system.new_pcoord_array(),
        )

        # Select the history to use
        # The following takes a random number in the interval 0 <= x < glom.weight, then
        # sees where this value falls among the (sorted) weights of the segments being merged;
        # this ensures that a walker with (e.g.) twice the weight of its brethren has twice the
        # probability of having its history selected for continuation
        iparent = np.digitize((random.uniform(0, glom.weight),), cumul_weight)[0]
        gparent_seg = segments[iparent]

        # Inherit history from this segment ("gparent" stands for "glom parent", as opposed to historical
        # parent).
        glom.parent_id = gparent_seg.parent_id
        glom.pcoord[0, :] = gparent_seg.pcoord[0, :]

        # Weight comes from all segments being merged, and therefore all their
        # parent segments
        glom.wtg_parent_ids = set()
        for segment in segments:
            glom.wtg_parent_ids |= segment.wtg_parent_ids

        # The historical parent of gparent is continued; all others are marked as merged
        for segment in segments:
            if segment is gparent_seg:
                # we must ignore initial states here...
                if segment.parent_id >= 0:
                    self._parent_map[segment.parent_id].endpoint_type = Segment.SEG_ENDPOINT_CONTINUES
            else:
                # and "unuse" an initial state here (recall that initial states are in 1:1 correspondence
                # with the segments they initiate), except when a previously-split particle is being
                # merged
                if segment.parent_id >= 0:
                    self._parent_map[segment.parent_id].endpoint_type = Segment.SEG_ENDPOINT_MERGED
                else:
                    if segment.initial_state_id in {segment.initial_state_id for segment in bin}:
                        log.debug('initial state in use by other walker; not removing')
                    else:
                        initial_state = self.used_initial_states.pop(segment.initial_state_id)
                        log.debug('freeing initial state {!r} for future use (merged)'.format(initial_state))
                        self.avail_initial_states[initial_state.state_id] = initial_state
                        initial_state.iter_used = None

        if log.isEnabledFor(logging.DEBUG):
            log.debug('merging ({:d}) {!r} into 1:\n    {!r}'.format(len(segments), segments, glom))

        return glom, gparent_seg

    def _split_by_weight(self, bin, target_count, ideal_weight):
        '''Split overweight particles'''

        segments = np.array(sorted(bin, key=operator.attrgetter('weight')), dtype=np.object_)
        weights = np.array(list(map(operator.attrgetter('weight'), segments)))

        if len(bin) > 0:
            assert target_count > 0

        to_split = segments[weights > self.weight_split_threshold * ideal_weight]

        for segment in to_split:
            m = int(math.ceil(segment.weight / ideal_weight))
            bin.remove(segment)
            new_segments_list = self._split_walker(segment, m, bin)
            bin.update(new_segments_list)

    def _merge_by_weight(self, bin, target_count, ideal_weight):
        '''Merge underweight particles'''

        while True:
            segments = np.array(sorted(bin, key=operator.attrgetter('weight')), dtype=np.object_)
            weights = np.array(list(map(operator.attrgetter('weight'), segments)))
            cumul_weight = np.add.accumulate(weights)

            to_merge = segments[cumul_weight <= ideal_weight * self.weight_merge_cutoff]
            if len(to_merge) < 2:
                return
            bin.difference_update(to_merge)
            new_segment, parent = self._merge_walkers(to_merge, cumul_weight, bin)
            bin.add(new_segment)

    def _adjust_count(self, bin, subgroups, target_count):
        weight_getter = operator.attrgetter('weight')
        # Order subgroups by the sum of their weights.
        if len(subgroups) > target_count:
            sorted_subgroups = [set()]
            for i in bin:
                sorted_subgroups[0].add(i)
        else:
            sorted_subgroups = sorted(subgroups, key=lambda gp: sum(seg.weight for seg in gp))
        # Loops over the groups, splitting/merging until the proper count has been reached.  This way, no trajectories are accidentally destroyed.

        # split
        while len(bin) < target_count:
            for i in sorted_subgroups:
                log.debug('adjusting counts by splitting')
                # always split the highest probability walker into two
                segments = sorted(i, key=weight_getter)
                bin.remove(segments[-1])
                i.remove(segments[-1])
                new_segments_list = self._split_walker(segments[-1], 2, bin)
                i.update(new_segments_list)
                bin.update(new_segments_list)

                if len(bin) == target_count:
                    break

        # merge
        while len(bin) > target_count:
            sorted_subgroups.reverse()
            # Adjust to go from lowest weight group to highest to merge
            for i in sorted_subgroups:
                # Ensures that there are least two walkers to merge
                if len(i) > 1:
                    log.debug('adjusting counts by merging')
                    # always merge the two lowest-probability walkers
                    segments = sorted(i, key=weight_getter)
                    bin.difference_update(segments[:2])
                    i.difference_update(segments[:2])
                    merged_segment, parent = self._merge_walkers(segments[:2], cumul_weight=None, bin=bin)
                    i.add(merged_segment)
                    bin.add(merged_segment)

                    # As long as we're changing the merge_walkers and split_walkers, adjust them so that they don't update the bin within the function
                    # and instead update the bin here.  Assuming nothing else relies on those.  Make sure with grin.
                    # in bash, "find . -name \*.py | xargs fgrep -n '_merge_walkers'"
                    if len(bin) == target_count:
                        break

    def _merge_by_threshold(self, bin, subgroup):
        # merge to satisfy weight thresholds
        # this gets rid of weights that are too small
        while True:
            segments = np.array(sorted(subgroup, key=operator.attrgetter('weight')), dtype=np.object_)
            weights = np.array(list(map(operator.attrgetter('weight'), segments)))
            cumul_weight = np.add.accumulate(weights)

            to_merge = segments[weights < self.smallest_allowed_weight]
            if len(to_merge) < 2:
                return
            bin.difference_update(to_merge)
            subgroup.difference_update(to_merge)
            new_segment, parent = self._merge_walkers(to_merge, cumul_weight, bin)
            bin.add(new_segment)
            subgroup.add(new_segment)

    def _split_by_threshold(self, bin, subgroup):
        # split to satisfy weight thresholds
        # this splits walkers that are too big
        segments = np.array(sorted(subgroup, key=operator.attrgetter('weight')), dtype=np.object_)
        weights = np.array(list(map(operator.attrgetter('weight'), segments)))

        to_split = segments[weights > self.largest_allowed_weight]
        for segment in to_split:
            m = int(math.ceil(segment.weight / self.largest_allowed_weight))
            bin.remove(segment)
            subgroup.remove(segment)
            new_segments_list = self._split_walker(segment, m, bin)
            bin.update(new_segments_list)
            subgroup.update(new_segments_list)

    def _check_pre(self):
        for ibin, _bin in enumerate(self.next_iter_binning):
            if self.bin_target_counts[ibin] == 0 and len(_bin) > 0:
                raise ConsistencyError('bin {:d} has target count of 0 but contains {:d} walkers'.format(ibin, len(_bin)))

    def _check_post(self):
        for segment in self.next_iter_segments:
            if segment.weight == 0:
                raise ConsistencyError('segment {!r} has weight of zero')

    def _prep_we(self):
        '''Prepare internal state for WE recycle/split/merge.'''
        self._parent_map = {}
        self.next_iter_binning = self.bin_mapper.construct_bins()

    def _run_we(self):
        '''Run recycle/split/merge. Do not call this function directly; instead, use
        populate_initial(), rebin_current(), or construct_next().'''
        self._recycle_walkers()

        # sanity check
        self._check_pre()

        # Regardless of current particle count, always split overweight particles and merge underweight particles
        # Then and only then adjust for correct particle count
        total_number_of_subgroups = 0
        total_number_of_particles = 0
        for (ibin, bin) in enumerate(self.next_iter_binning):
            if len(bin) == 0:
                continue

            # Splits the bin into subgroups as defined by the called function
            target_count = self.bin_target_counts[ibin]
            subgroups = self.subgroup_function(self, ibin, **self.subgroup_function_kwargs)
            total_number_of_subgroups += len(subgroups)
            # Clear the bin
            segments = np.array(sorted(bin, key=operator.attrgetter('weight')), dtype=np.object_)
            weights = np.array(list(map(operator.attrgetter('weight'), segments)))
            ideal_weight = weights.sum() / target_count
            bin.clear()
            # Determines to see whether we have more sub bins than we have target walkers in a bin (or equal to), and then uses
            # different logic to deal with those cases.  Should devolve to the Huber/Kim algorithm in the case of few subgroups.
            if len(subgroups) >= target_count:
                for i in subgroups:
                    # Merges all members of set i.  Checks to see whether there are any to merge.
                    if len(i) > 1:
                        (segment, parent) = self._merge_walkers(
                            list(i),
                            np.add.accumulate(np.array(list(map(operator.attrgetter('weight'), i)))),
                            i,
                        )
                        i.clear()
                        i.add(segment)
                    # Add all members of the set i to the bin.  This keeps the bins in sync for the adjustment step.
                    bin.update(i)

                if len(subgroups) > target_count:
                    self._adjust_count(bin, subgroups, target_count)

            if len(subgroups) < target_count:
                for i in subgroups:
                    self._split_by_weight(i, target_count, ideal_weight)
                    self._merge_by_weight(i, target_count, ideal_weight)
                    # Same logic here.
                    bin.update(i)
                if self.do_adjust_counts:
                    # A modified adjustment routine is necessary to ensure we don't unnecessarily destroy trajectory pathways.
                    self._adjust_count(bin, subgroups, target_count)
            if self.do_thresholds:
                for i in subgroups:
                    self._split_by_threshold(bin, i)
                    self._merge_by_threshold(bin, i)
                for iseg in bin:
                    if iseg.weight > self.largest_allowed_weight or iseg.weight < self.smallest_allowed_weight:
                        log.warning(
                            f'Unable to fulfill threshold conditions for {iseg}. The given threshold range is likely too small.'
                        )
            total_number_of_particles += len(bin)
        log.debug('Total number of subgroups: {!r}'.format(total_number_of_subgroups))

        self._check_post()

        self.new_weights = self.new_weights or []

        log.debug('used initial states: {!r}'.format(self.used_initial_states))
        log.debug('available initial states: {!r}'.format(self.avail_initial_states))

    def populate_initial(self, initial_states, weights, system=None):
        '''Create walkers for a new weighted ensemble simulation.

        One segment is created for each provided initial state, then binned and split/merged
        as necessary. After this function is called, next_iter_segments will yield the new
        segments to create, used_initial_states will contain data about which of the
        provided initial states were used, and avail_initial_states will contain data about
        which initial states were unused (because their corresponding walkers were merged
        out of existence).
        '''

        # This has to be down here to avoid an import race
        from westpa.core.data_manager import weight_dtype

        EPS = np.finfo(weight_dtype).eps

        system = system or westpa.rc.get_system_driver()
        self.new_iteration(
            initial_states=[], target_states=[], bin_mapper=system.bin_mapper, bin_target_counts=system.bin_target_counts
        )

        # Create dummy segments
        segments = []
        for (seg_id, (initial_state, weight)) in enumerate(zip(initial_states, weights)):
            dummy_segment = Segment(
                n_iter=0,
                seg_id=seg_id,
                parent_id=-(initial_state.state_id + 1),
                weight=weight,
                wtg_parent_ids=set([-(initial_state.state_id + 1)]),
                pcoord=system.new_pcoord_array(),
                status=Segment.SEG_STATUS_PREPARED,
            )
            dummy_segment.pcoord[[0, -1]] = initial_state.pcoord
            segments.append(dummy_segment)

        # Adjust weights, if necessary
        tprob = sum(weights)
        if abs(1.0 - tprob) > len(weights) * EPS:
            pscale = 1.0 / tprob
            log.warning('Weights of initial segments do not sum to unity; scaling by {:g}'.format(pscale))
            for segment in segments:
                segment.weight *= pscale

        self.assign(segments, initializing=True)
        self.construct_next()

        # We now have properly-constructed initial segments, except for parent information,
        # and we need to  mark initial states as used or unused
        istates_by_id = {state.state_id: state for state in initial_states}
        dummysegs_by_id = self._parent_map

        # Don't add start states to the list of available initial states.
        # They're only meant to be used in the first iteration, so nothing should ever be recycled into them.
        # Thus, they're not available.
        self.avail_initial_states = {
            k: v for (k, v) in istates_by_id.items() if not v.istate_type == InitialState.ISTATE_TYPE_START
        }

        for state in self.avail_initial_states.keys():
            if self.avail_initial_states[state].istate_type == InitialState.ISTATE_TYPE_START:
                self.avail_initial_states.pop(state)

        self.used_initial_states = {}
        for segment in self.next_iter_segments:
            segment.parent_id = dummysegs_by_id[segment.parent_id].parent_id
            segment.wtg_parent_ids = set([segment.parent_id])
            assert segment.initpoint_type == Segment.SEG_INITPOINT_NEWTRAJ
            istate = istates_by_id[segment.initial_state_id]
            try:
                self.used_initial_states[istate.state_id] = self.avail_initial_states.pop(istate.state_id)
            except KeyError:
                # Shared by more than one segment, and already marked as used
                pass

        for used_istate in self.used_initial_states.values():
            used_istate.iter_used = 1

    def rebin_current(self, parent_segments):
        '''Reconstruct walkers for the current iteration based on (presumably) new binning.
        The previous iteration's segments must be provided (as ``parent_segments``) in order
        to update endpoint types appropriately.'''

        self._prep_we()
        self._parent_map = {segment.seg_id: segment for segment in parent_segments}

        # Create new segments for the next iteration
        # We assume that everything is going to continue without being touched by recycling or WE, and
        # adjust later
        new_pcoord_array = self.system.new_pcoord_array
        n_iter = None

        for ibin, _bin in enumerate(self.final_binning):
            for segment in _bin:
                if n_iter is None:
                    n_iter = segment.n_iter
                else:
                    assert segment.n_iter == n_iter

                new_segment = Segment(
                    n_iter=segment.n_iter,
                    parent_id=segment.parent_id,
                    weight=segment.weight,
                    wtg_parent_ids=set(segment.wtg_parent_ids or []),
                    pcoord=new_pcoord_array(),
                    status=Segment.SEG_STATUS_PREPARED,
                )
                new_segment.pcoord[0] = segment.pcoord[0]
                self.next_iter_binning[ibin].add(new_segment)

        self._run_we()

    def construct_next(self):
        '''Construct walkers for the next iteration, by running weighted ensemble recycling
        and bin/split/merge on the segments previously assigned to bins using ``assign``.
        Enough unused initial states must be present in ``self.avail_initial_states`` for every recycled
        walker to be assigned an initial state.

        After this function completes, ``self.flux_matrix`` contains a valid flux matrix for this
        iteration (including any contributions from recycling from the previous iteration), and
        ``self.next_iter_segments`` contains a list of segments ready for the next iteration,
        with appropriate values set for weight, endpoint type, parent walkers, and so on.
        '''

        self._prep_we()

        # Create new segments for the next iteration
        # We assume that everything is going to continue without being touched by recycling or WE, and
        # adjust later
        new_pcoord_array = self.system.new_pcoord_array
        n_iter = None

        for ibin, _bin in enumerate(self.final_binning):
            for segment in _bin:
                if n_iter is None:
                    n_iter = segment.n_iter
                else:
                    assert segment.n_iter == n_iter

                segment.endpoint_type = Segment.SEG_ENDPOINT_CONTINUES
                new_segment = Segment(
                    n_iter=segment.n_iter + 1,
                    parent_id=segment.seg_id,
                    weight=segment.weight,
                    wtg_parent_ids=[segment.seg_id],
                    pcoord=new_pcoord_array(),
                    status=Segment.SEG_STATUS_PREPARED,
                )
                new_segment.pcoord[0] = segment.pcoord[-1]
                self.next_iter_binning[ibin].add(new_segment)

                # Store a link to the parent segment, so we can update its endpoint status as we need,
                # based on its ID
                self._parent_map[segment.seg_id] = segment

        self._run_we()

        log.debug('used initial states: {!r}'.format(self.used_initial_states))
        log.debug('available initial states: {!r}'.format(self.avail_initial_states))

    def _log_bin_stats(self, bin, heading=None, level=logging.DEBUG):
        if log.isEnabledFor(level):
            weights = sorted(np.array(list(map(operator.attrgetter('weight'), bin))))
            bin_label = getattr(bin, 'label', None) or ''
            log_fmt = '\n      '.join(
                [
                    '',
                    'stats for bin {bin_label!r} {heading}',
                    '  count: {bin.count:d}, target count: {bin.target_count:d}',
                    '  total weight: {bin.weight:{weight_spec}},  ideal weight: {ideal_weight:{weight_spec}}',
                    '  mean weight:  {mean_weight:{weight_spec}},  stdev weight: {stdev_weight:{weight_spec}}',
                    '  min weight:   {min_weight:{weight_spec}},  med weight  : {median_weight:{weight_spec}}'
                    + ', max weight: {max_weight:{weight_spec}}',
                ]
            )
            log_msg = log_fmt.format(
                log_fmt,
                weight_spec='<12.6e',
                bin_label=bin_label,
                heading=heading,
                bin=bin,
                ideal_weight=bin.weight / bin.target_count,
                mean_weight=weights.mean(),
                stdev_weight=weights.std(),
                min_weight=weights[0],
                median_weight=np.median(weights),
                max_weight=weights[-1],
            )
            log.log(level, log_msg)



import numpy as np
import os

from mdtraj import Trajectory, load as load_traj, FormatRegistry, formats as mdformats
from mdtraj.core.trajectory import _TOPOLOGY_EXTS, _get_extension as get_extension

FormatRegistry.loaders['.rst'] = mdformats.amberrst.load_restrt
FormatRegistry.fileobjects['.rst'] = mdformats.AmberRestartFile

TRAJECTORY_EXTS = list(FormatRegistry.loaders.keys())
TOPOLOGY_EXTS = list(_TOPOLOGY_EXTS)
for ext in [".h5", ".hdf5", ".lh5"]:
    TOPOLOGY_EXTS.remove(ext)


class WESTTrajectory(Trajectory):
    '''A subclass of ``mdtraj.Trajectory`` that contains the trajectory of atom coordinates with
    pointers denoting the iteration number and segment index of each frame.'''

    def __init__(
        self,
        coordinates,
        topology=None,
        time=None,
        iter_labels=None,
        seg_labels=None,
        pcoords=None,
        parent_ids=None,
        unitcell_lengths=None,
        unitcell_angles=None,
    ):
        if isinstance(coordinates, Trajectory):
            xyz = coordinates.xyz
            topology = coordinates.topology if topology is None else topology
            time = coordinates.time if time is None else time
            unitcell_lengths = coordinates.unitcell_lengths if unitcell_lengths is None else unitcell_lengths
            unitcell_angles = coordinates.unitcell_angles if unitcell_angles is None else unitcell_angles
        else:
            xyz = coordinates

        super(WESTTrajectory, self).__init__(xyz, topology, time, unitcell_lengths, unitcell_angles)
        self._shape = None
        self.iter_labels = iter_labels
        self.seg_labels = seg_labels
        self.pcoords = pcoords
        self.parent_ids = parent_ids

    def _string_summary_basic(self):
        """Basic summary of WESTTrajectory in string form."""
        unitcell_str = 'and unitcells' if self._have_unitcell else 'without unitcells'
        value = "%s with %d frames, %d atoms, %d residues, %s" % (
            self.__class__.__name__,
            self.n_frames,
            self.n_atoms,
            self.n_residues,
            unitcell_str,
        )
        return value

    def _check_labels(self, value):
        if value is None:
            value = 0
        elif isinstance(value, list):
            value = np.array(value)

        if np.isscalar(value):
            value = np.array([value] * self.n_frames, dtype=int)
        elif value.shape != (self.n_frames,):
            raise ValueError('Wrong shape. Got %s, should be %s' % (value.shape, (self.n_frames,)))

        return value

    def _check_pcoords(self, value):
        if value is None:
            value = 0.0
        elif isinstance(value, list):
            value = np.array(value)

        if np.isscalar(value):
            value = np.array([(value,)] * self.n_frames, dtype=float)

        if value.ndim == 1:
            value = np.tile(value, (self.n_frames, 1))
        elif value.ndim != 2:
            raise ValueError('pcoords must be a 2-D array')

        elif value.shape[0] != self.n_frames:
            raise ValueError('Wrong length. Got %s, should be %s' % (value.shape[0], self.n_frames))

        return value

    def iter_label_values(self):
        visited_ids = []

        for i in self.iter_labels:
            if i in visited_ids:
                continue
            visited_ids.append(i)
            yield i

    def seg_label_values(self, iteration=None):
        seg_labels = self.seg_labels[self.iter_labels == iteration]
        visited_ids = []

        for j in seg_labels:
            if j in visited_ids:
                continue
            visited_ids.append(j)
            yield j

    @property
    def label_values(self):
        for i in self.iter_label_values():
            for j in self.seg_label_values(i):
                yield i, j

    def _iter_blocks(self):
        for i, j in self.label_values:
            IandJ = np.logical_and(self.iter_labels == i, self.seg_labels == j)
            yield i, j, IandJ

    @property
    def iter_labels(self):
        """Iteration index corresponding to each frame

        Returns
        -------
        time : np.ndarray, shape=(n_frames,)
            The iteration index corresponding to each frame
        """
        return self._iters

    @iter_labels.setter
    def iter_labels(self, value):
        "Set the iteration index corresponding to each frame"

        self._iters = self._check_labels(value)
        self._shape = None

    @property
    def seg_labels(self):
        """Segment index corresponding to each frame

        Returns
        -------
        time : np.ndarray, shape=(n_frames,)
            The segment index corresponding to each frame
        """
        return self._segs

    @seg_labels.setter
    def seg_labels(self, value):
        "Set the segment index corresponding to each frame"

        self._segs = self._check_labels(value)
        self._shape = None

    @property
    def pcoords(self):
        return self._pcoords

    @pcoords.setter
    def pcoords(self, value):
        self._pcoords = self._check_pcoords(value)

    @property
    def parent_ids(self):
        return self._parent_ids

    @parent_ids.setter
    def parent_ids(self, value):
        self._parent_ids = self._check_labels(value)

    def join(self, other, check_topology=True, discard_overlapping_frames=False):
        """Join two ``Trajectory``s. This overrides ``mdtraj.Trajectory.join``
        so that it also handles WESTPA pointers.
        ``mdtraj.Trajectory.join``'s documentation for more details.
        """
        if isinstance(other, Trajectory):
            other = [other]

        new_traj = super(WESTTrajectory, self).join(
            other, check_topology=check_topology, discard_overlapping_frames=discard_overlapping_frames
        )

        trajectories = [self] + other
        if discard_overlapping_frames:
            for i in range(len(trajectories) - 1):
                x0 = trajectories[i].xyz[-1]
                x1 = trajectories[i + 1].xyz[0]

                if np.all(np.abs(x1 - x0) < 2e-3):
                    trajectories[i] = trajectories[i][:-1]

        iter_labels = []
        seg_labels = []
        parent_ids = []
        pshape = self.pcoords.shape
        pcoords = []

        for t in trajectories:
            if hasattr(t, "iter_labels"):
                iters = t.iter_labels
            else:
                iters = np.zeros(len(t)) - 1  # default iter label: -1

            iter_labels.append(iters)

            if hasattr(t, "seg_labels"):
                segs = t.seg_labels
            else:
                segs = np.zeros(len(t)) - 1  # default seg label: -1

            seg_labels.append(segs)

            if hasattr(t, "parent_ids"):
                pids = t.parent_ids
            else:
                pids = np.zeros(len(t)) - 1  # default parent_id: -1

            parent_ids.append(pids)

            if hasattr(t, "pcoords"):
                p = t.pcoords
            else:
                p = np.zeros((len(t), pshape[-1]), dtype=float)  # default pcoord: 0.0

            pcoords.append(p)

        iter_labels = np.concatenate(iter_labels)
        seg_labels = np.concatenate(seg_labels)
        parent_ids = np.concatenate(parent_ids)
        pcoords = np.concatenate(pcoords)

        new_westpa_traj = WESTTrajectory(
            new_traj, iter_labels=iter_labels, seg_labels=seg_labels, pcoords=pcoords, parent_ids=parent_ids
        )

        return new_westpa_traj

    def slice(self, key, copy=True):
        """Slice the ``Trajectory``. This overrides ``mdtraj.Trajectory.slice``
        so that it also handles WESTPA pointers. Please see
        ``mdtraj.Trajectory.slice``'s documentation for more details.
        """

        if isinstance(key, tuple):
            if self._shape is None:
                uniq_iters = np.unique(self.iter_labels)
                max_iter = uniq_iters.max()
                max_seg = self.seg_labels.max()
                max_n_trajs = 0
                for _, _, block in self._iter_blocks():
                    n_trajs = block.sum()
                    if n_trajs > max_n_trajs:
                        max_n_trajs = n_trajs

                self._shape = (max_iter, max_seg, max_n_trajs)
            else:
                max_iter, max_seg, max_n_trajs = self._shape

            M = np.full((max_iter + 1, max_seg + 1, max_n_trajs), -1, dtype=int)
            all_traj_indices = np.arange(self.n_frames, dtype=int)
            for i, j, block in self._iter_blocks():
                traj_indices = all_traj_indices[block]

                for k, traj_idx in enumerate(traj_indices):
                    M[i, j, k] = traj_idx

            selected_indices = M[key].flatten()
            if np.isscalar(selected_indices):
                selected_indices = np.array([selected_indices])
            key = selected_indices[selected_indices != -1]

        iters = self.iter_labels[key]
        segs = self.seg_labels[key]
        pcoords = self.pcoords[key, :]
        parent_ids = self.parent_ids[key]

        traj = super(WESTTrajectory, self).slice(key, copy)
        traj.iter_labels = iters
        traj.seg_labels = segs
        traj.pcoords = pcoords
        traj.parent_ids = parent_ids

        return traj


def load_trajectory(folder):
    '''Load trajectory from ``folder`` using ``mdtraj`` and return a ``mdtraj.Trajectory``
    object. The folder should contain a trajectory and a topology file (with a recognizable
    extension) that is supported by ``mdtraj``. The topology file is optional if the
    trajectory file contains topology data (e.g., HDF5 format).
    '''
    traj_file = top_file = None
    for filename in os.listdir(folder):
        filepath = os.path.join(folder, filename)
        if not os.path.isfile(filepath):
            continue

        ext = get_extension(filename).lower()
        if ext in TOPOLOGY_EXTS and top_file is None:
            top_file = filename
        elif ext in TRAJECTORY_EXTS and traj_file is None:
            traj_file = filename

        if top_file is not None and traj_file is not None:
            break

    if traj_file is None:
        raise ValueError('trajectory file not found')

    traj_file = os.path.join(folder, traj_file)

    kwargs = {}
    if top_file is not None:
        top_file = os.path.join(folder, top_file)
        kwargs['top'] = top_file

    traj = load_traj(traj_file, **kwargs)
    return traj



import logging

import numpy as np

import westpa
from westpa.core.binning import NopMapper

log = logging.getLogger(__name__)


class WESTSystem:
    '''A description of the system being simulated, including the dimensionality and
    data type of the progress coordinate, the number of progress coordinate entries
    expected from each segment, and binning. To construct a simulation, the user must
    subclass WESTSystem and set several instance variables.

    At a minimum, the user must subclass ``WESTSystem`` and override
    :method:`initialize` to set the data type and dimensionality of progress
    coordinate data and define a bin mapper.

    :ivar pcoord_ndim:    The number of dimensions in the progress coordinate.
                          Defaults to 1 (i.e. a one-dimensional progress
                          coordinate).
    :ivar pcoord_dtype:   The data type of the progress coordinate, which must be
                          callable (e.g. ``np.float32`` and ``long`` will work,
                          but ``'<f4'`` and ``'<i8'`` will not).  Defaults to
                          ``np.float64``.
    :ivar pcoord_len:     The length of the progress coordinate time series
                          generated by each segment, including *both* the initial
                          and final values.  Defaults to 2 (i.e. only the initial
                          and final progress coordinate values for a segment are
                          returned from propagation).
    :ivar bin_mapper:     A bin mapper describing the progress coordinate space.
    :ivar bin_target_counts: A vector of target counts, one per bin.
    '''

    def __init__(self, rc=None):
        self.rc = rc or westpa.rc

        # Number of dimentions in progress coordinate data
        self.pcoord_ndim = 1

        # Length of progress coordinate data for each segment
        self.pcoord_len = 2

        # Data type of progress coordinate
        self.pcoord_dtype = np.float32

        # Mapper
        self.bin_mapper = NopMapper()
        self._bin_target_counts = None

        self.bin_target_counts = [1]

    @property
    def bin_target_counts(self):
        return self._bin_target_counts

    @bin_target_counts.setter
    def bin_target_counts(self, target_counts):
        maxcount = max(target_counts)
        self._bin_target_counts = np.array(target_counts, dtype=np.min_scalar_type(maxcount))

    def initialize(self):
        '''Prepare this system object for use in simulation or analysis,
        creating a bin space, setting replicas per bin, and so on. This
        function is called whenever a WEST tool creates an instance of the
        system driver.
        '''
        pass

    def prepare_run(self):
        '''Prepare this system for use in a simulation run. Called by w_run in
        all worker processes.'''
        pass

    def finalize_run(self):
        '''A hook for system-specific processing for the end of a simulation run
        (as defined by such things as maximum wallclock time, rather than perhaps
        more scientifically-significant definitions of "the end of a simulation run")'''
        pass

    def new_pcoord_array(self, pcoord_len=None):
        '''Return an appropriately-sized and -typed pcoord array for a timepoint, segment,
        or number of segments. If ``pcoord_len`` is not specified (or None), then
        a length appropriate for a segment is returned.'''

        if pcoord_len is None:
            pcoord_len = self.pcoord_len
        return np.zeros((pcoord_len, self.pcoord_ndim), self.pcoord_dtype)

    def new_region_set(self):
        raise NotImplementedError('This method has been removed.')



import westpa
from westpa.core.progress import ProgressIndicator
from westpa.tools.core import WESTToolComponent


class ProgressIndicatorComponent(WESTToolComponent):
    def __init__(self):
        super().__init__()
        self.indicator = None

    def add_args(self, parser):
        pass

    def process_args(self, args):
        self.indicator = ProgressIndicator()
        if westpa.rc.quiet_mode or westpa.rc.verbose_mode or westpa.rc.debug_mode:
            self.indicator.fancy = False
